/*
 *	Copyright (C) 1997 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: daui.c,v 1.6 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <assert.h>
#if defined HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#if defined HAVE_MACH_O_DYLD_H
#include <mach-o/dyld.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/daui.h>

#include "dauiP.h"

#define	PADSIZE		(sizeof (int))

#define	PAGESIZE	8192

struct _c {
	char *path;
	FILE *f;
};

struct _f {
	void *p;
	size_t len;
	struct daui_hdr h, *hp;
	struct au_index_t *index;

	char dirname[MAXPATHLEN];
	char handle[MAXPATHLEN];
	char abase[MAXPATHLEN];

	struct _c cache;
};

struct _e {
	void *cw;

	char handle[MAXPATHLEN];
	char cmd[MAXPATHLEN];
};

struct _d {
	void *cw;

	void *dl;
	char handle[MAXPATHLEN];
	char path[MAXPATHLEN];
	char name[MAXPATHLEN];
	ssize_t (*func)(char const *, char const *, char **);
};

struct dw_FILE {
	int magic;
	union {
		struct _f f;
		struct _e e;
		struct _d d;
	} u;
	char *pool;
	size_t size;
};

/* extern char **environ; */

static ssize_t dau_f(FILE *, size_t, size_t, char **, size_t *);
static ssize_t dau_e(char *, char *[], char **, size_t *);
static ssize_t dau_d(ssize_t (*)(char const *, char const *, char **), char const *, char const *, char **, size_t *);

static void dau_ccln(DWFILE *);
static void dau_crst(DWFILE *);
static FILE *dau_fopen(DWFILE *, char const *);

DWFILE *
dau_open(dirname, handle, abase)
char const *dirname, *handle, *abase;
{
	int d = -1;
	DWFILE *f = NULL;
	struct stat sb;
	struct _f *p;
	char path[MAXPATHLEN];

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}
	p = &f->u.f;
	p->p = MAP_FAILED;
	p->len = 0;
	f->pool = NULL;
	f->size = 0;
	dau_ccln(f);

	strlcpy(p->dirname, dirname, sizeof p->dirname);
	strlcpy(p->handle, handle, sizeof p->handle);
	strlcpy(p->abase, abase, sizeof p->abase);

	dw_mkauipath(path, sizeof path, dirname, handle, abase, -1, 0);

	if ((d = open(path, O_RDONLY))==-1) {
		dwerrno = errno;
		goto err;
	}

	if (fstat(d, &sb)==-1) {
		dwerrno = errno;
		goto err;
	}

	p->len = sb.st_size;
	if ((p->p = mmap(0, p->len, PROT_READ, MAP_SHARED, d, 0))==MAP_FAILED) {
		dwerrno = errno;
		goto err;
	}

	close(d);
	d = -1;

	p->hp = p->p;
	memmove(&p->h, p->hp, sizeof p->h);

	p->index = (struct au_index_t *)(p->hp+1);
	f->magic = DAUI_MAGIC_F;

	return f;
err:
	if (d!=-1) {
		close(d);
	}
	dau_close(f);
	return NULL;
}

DWFILE *
dau_open_cmd(handle, cmd, cw)
char const *handle, *cmd;
void *cw;
{
	DWFILE *f = NULL;
	struct _e *p;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}
	p = &f->u.e;
	f->pool = NULL;
	f->size = 0;

	strlcpy(p->handle, handle, sizeof p->handle);
	strlcpy(p->cmd, cmd, sizeof p->cmd);

	p->cw = cw;
	f->magic = DAUI_MAGIC_E;

	return f;
}

DWFILE *
dau_open_dl(handle, path, name, cw, mode)
char const *handle, *path, *name;
void *cw;
{
	DWFILE *f = NULL;
	struct _d *p;
	ssize_t (*v)(char const *, char const *, char **);

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}
	p = &f->u.d;
	f->pool = NULL;
	f->size = 0;

	strlcpy(p->handle, handle, sizeof p->handle);
	strlcpy(p->path, path, sizeof p->path);
	strlcpy(p->name, name, sizeof p->name);

#if defined HAVE_MACH_O_DYLD_H
	p->dl = NULL;
	if (_dyld_present()==0) {
		free(f);
		perror("_dyld_present");
		return NULL;
	}
	_dyld_lookup_and_bind(p->name, (unsigned long *)&v, NULL);
#else
#if HAVE_DLFCN_H
	if (!(p->dl = dlopen(p->path, mode))) {
		free(f);
		dwerrno = errno;
		return NULL;
	}

	if (!(v = dlsym(p->dl, p->name))) {
		free(f);
		dwerrno = errno;
		return NULL;
	}
#else
	fprintf(stderr, "lddll: dynamic linker is not configured.\n");
	free(f);
	return NULL;
#endif
#endif
	p->func = v;

	p->cw = cw;
	f->magic = DAUI_MAGIC_D;

	return f;
}

int
dau_close(f)
DWFILE *f;
{
	switch (f->magic) {
	case DAUI_MAGIC_F: {
			struct _f *p;
			p = &f->u.f;
			if (p->p!=MAP_FAILED) {
				munmap(p->p, p->len);
			}
		}
		break;
	case DAUI_MAGIC_E: {
			struct _e *p;
			p = &f->u.e;
		}
		break;
	case DAUI_MAGIC_D: {
#if defined HAVE_MACH_O_DYLD_H
			/* do nothing */
#else
#if HAVE_DLFCN_H
			struct _d *p;
			p = &f->u.d;
			dlclose(p->dl);
#endif
#endif
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (f->pool && f->size>PADSIZE) {
		f->pool[PADSIZE] = '\0';
	}
	free(f->pool);
	f->magic = 0;
	dau_crst(f);
	free(f);
	return 0;
}

ssize_t
dau_nentries(f)
DWFILE *f;
{
	switch (f->magic) {
	case DAUI_MAGIC_F: {
			struct _f *p;
			p = &f->u.f;
			return p->h.nentries;
		}
		break;
	case DAUI_MAGIC_E: {
			struct _e *p;
			p = &f->u.e;
			if (p->cw) {
				return dcw_size(p->cw);
			}
		}
		break;
	case DAUI_MAGIC_D: {
			struct _d *p;
			p = &f->u.d;
			if (p->cw) {
				return dcw_size(p->cw);
			}
		}
		break;
	default:
		break;
	}
	dwerrno = DWE_INVAL;
	return -1;
}

ssize_t
dau_gets(f, id, p)
DWFILE *f;
idx_t id;
char const **p;
{
	ssize_t r = -1;

	switch (f->magic) {
	case DAUI_MAGIC_F: {
			FILE *g;
			ssize_t frag;
			size_t addr;
			size_t alen;
			char path[MAXPATHLEN];
			struct _f *p;

			p = &f->u.f;

			if (id==0 || p->h.nentries<id) {
				dwerrno = DWE_BADR;
				return -1;
			}

			frag = p->index[id].fragment;
			addr = p->index[id].offset;
			alen = p->index[id].length;

			dw_mkauipath(path, sizeof path, p->dirname, p->handle, p->abase, frag, 0);

			if (!(g = dau_fopen(f, path))) {
				dwerrno = errno;
				return -1;
			}

			r = dau_f(g, addr, alen, &f->pool, &f->size);
			assert(r==-1 || r==alen);
		}
		break;
	case DAUI_MAGIC_E: {
			char *argv[4];
			char const *key;
			struct _e *p;
			char keys[32];

			p = &f->u.e;

			if (p->cw) {
				if (!(key = dcw_id2name(p->cw, id))) {
					return -1;
				}
			}
			else {
				snprintf(keys, sizeof keys, "%s", idxtostr(id));
				key = keys;
			}

			argv[0] = p->cmd;
			argv[1] = p->handle;
			argv[2] = (char *)key;
			argv[3] = NULL;

			r = dau_e(p->cmd, argv, &f->pool, &f->size);
		}
		break;
	case DAUI_MAGIC_D: {
			char const *key;
			struct _d *p;
			char keys[32];

			p = &f->u.d;

			if (p->cw) {
				if (!(key = dcw_id2name(p->cw, id))) {
					return -1;
				}
			}
			else {
				snprintf(keys, sizeof keys, "%s", idxtostr(id));
				key = keys;
			}

			r = dau_d(p->func, p->handle, key, &f->pool, &f->size);
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (r!=-1) {
		f->pool[r+PADSIZE] = '\0';
		*p = f->pool+PADSIZE;
#if 0
{
size_t i;
#define	cc(c)	fprintf(stderr, ' '<=(c) && (c)<='~' ? "%c   " : "%03o ", (c))
for (i=-1; i<16; i++) {
	cc((*p)[i]);
}
fprintf(stderr, "\n");
for (i=r-16; i<r+1; i++) {
	cc((*p)[i]);
}
fprintf(stderr, "\n");
}
#endif

	}
	return r;
}

static ssize_t
dau_f(g, addr, alen, pool, size)
FILE *g;
size_t addr;
size_t alen;
char **pool;
size_t *size;
{
	size_t newsize;

	if ((newsize = alen+2*PADSIZE)>*size) {
		char *new;
		newsize = NALIGN(newsize, PAGESIZE);
		if (!(new = realloc(*pool, newsize))) {
			dwerrno = errno;
			return -1;
		}
		*pool = new;
		*size = newsize;
	}
	if (fseeko(g, addr, SEEK_SET)==-1) {
		dwerrno = errno;
		return -1;
	}
	if (fread(*pool+PADSIZE, 1, alen, g)!=alen) {
		dwerrno = errno;
		return -1;
	}
	return alen;
}

static ssize_t
dau_e(cmd, argv, pool, size)
char *cmd;
char *argv[];
char **pool;
size_t *size;
{
	int status;
	int fds[2];
	pid_t pid;
	ssize_t rrsize;
	char buf[1024];

	struct sigaction sa;

	if (pipe(fds)==-1) {
		dwerrno = errno;
		goto error;
	}
	if (sigaction(SIGCHLD, NULL, &sa)==-1) {
		sa.sa_handler = SIG_ERR;
	}
	signal(SIGCHLD, SIG_DFL);
	switch (pid = fork()) {
		ssize_t c;
	case -1:
		dwerrno = errno;
		goto error;
	case 0:
		close(fds[0]);
		if (fds[1]!=1) {
			dup2(fds[1], 1);
			close(fds[1]);
		}
		close(0);
#if 0
{
	size_t i;
	fprintf(stderr, "execve: %s", cmd);
	for (i=0; argv[i]; i++) {
		fprintf(stderr, " \"%s\"", argv[i]);
	}
	fprintf(stderr, "\n");
}
#endif
		execv(cmd, argv/*, environ*/);
		nperror(cmd);
		_exit(0);
	default:
		close(fds[1]);
		rrsize = 0;

		while ((c = read(fds[0], buf, sizeof buf))>=0) {
			size_t newsize;
			if ((newsize = rrsize+c+2*PADSIZE)>*size) {
				char *new;
				newsize = NALIGN(newsize, PAGESIZE);
				if (!(new = realloc(*pool, newsize))) {
					dwerrno = errno;
					goto error;
				}
				*pool = new;
				*size = newsize;
			}
			if (c==0) {
				/* at least, previous check for reallocation
				   should be done before braking this loop. */
				break;
			}
			memmove(*pool+PADSIZE+rrsize, buf, c);
			rrsize += c;
		}
		close(fds[0]);
		if (waitpid(pid, &status, 0)==-1) {
			goto error;
		}
		else if (WIFEXITED(status) && WEXITSTATUS(status)==0) {
			/* do nothing */
		}
		else {
			fprintf(stderr, "dau_e: %s ", cmd);
			if (WIFEXITED(status)) {
				fprintf(stderr, "exit %d", WEXITSTATUS(status));
			}
			else if (WIFSIGNALED(status)) {
				fprintf(stderr, "killed %d", WTERMSIG(status));
				if (WCOREDUMP(status)) {
					fprintf(stderr, "(core dumped)");
				}
			}
			if (WIFSTOPPED(status)) {
				fprintf(stderr, "stopped %d", WSTOPSIG(status));
			}
			fprintf(stderr, ".\n");
			goto error;
		}
	}
	if (sa.sa_handler!=SIG_ERR) {
		sigaction(SIGCHLD, &sa, NULL);
	}

	return rrsize;

error:
	if (sa.sa_handler!=SIG_ERR) {
		sigaction(SIGCHLD, &sa, NULL);
	}
	return -1;
}

static ssize_t
dau_d(func, handle, key, pool, size)
ssize_t (*func)(char const *, char const *, char **);
char const *handle, *key;
char **pool;
size_t *size;
{
	ssize_t alen;
	size_t newsize;
	char *buf;

	if (!func) {
		dwerrno = EINVAL;
		return -1;
	}
	if ((alen = (*func)(handle, key, &buf))==-1) {
		dwerrno = errno;
		return -1;
	}

	if ((newsize = alen+2*PADSIZE)>*size) {
		char *new;
		newsize = NALIGN(newsize, PAGESIZE);
		if (!(new = realloc(*pool, newsize))) {
			dwerrno = errno;
			return -1;
		}
		*pool = new;
		*size = newsize;
	}
	bcopy(buf, *pool+PADSIZE, alen);
	return alen;
}

static void
dau_ccln(f)
DWFILE *f;
{
	struct _c *c = &f->u.f.cache;
	c->path = NULL;
	c->f = NULL;
}

static void
dau_crst(f)
DWFILE *f;
{
	struct _c *c = &f->u.f.cache;
	if (c->f) {
		assert(c->path!=NULL);
		free(c->path);
		c->path = NULL;
		fclose(c->f);
		c->f = NULL;
	}
	else {
		assert(c->path==NULL);
	}
}

static FILE *
dau_fopen(f, path)
DWFILE *f;
char const *path;
{
	struct _c *c = &f->u.f.cache;
	if (c->path && !strcmp(c->path, path)) {
		assert(c->f);
		return c->f;
	}
	if (c->path) {
		assert(c->f);
		free(c->path);
		fclose(c->f);
	}
	else {
		assert(!c->f);
	}
	if (c->f = fopen(path, "r")) {
		if (c->path = strdup(path)) {
			return c->f;
		}
		else {
			fclose(c->f);
			c->f = NULL;
		}
	}
	else {
		c->path = NULL;
	}
	return NULL;
}
