/*
 *	Copyright (C) 1994, 1996, 1999
 *		Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dci.c,v 1.2 2007/03/16 01:33:12 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <netinet/in.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/nlog.h>
#include <geta/dci.h>

#include "dciP.h"

#if USE_CICONF

struct attr {
	char const *name;
	char const *value;
};

struct dataset {
	char const *handle;	/* mandatory */
	char const *shortname;	/* NULL */
	char const *dataroot;	/* supplied by user or automatically filled */
	u_int flags;		/* 0 */
	size_t nattrs, maxnattrs;
	struct attr *attrs;
};

static struct dataset datasets[NDSETS];
static char const *handles[NDSETS+1];
static struct dataset *hdatasets[NDSETS];
static struct dataset *sdatasets[NDSETS];

/* current dataset */
static struct dataset *c = NULL;
static size_t ndsets = 0;

static int set_dictionary(char const *);
static int conflictp(char const *, size_t, char const *, size_t);
static int ucompar(void const *, void const *);
static int hcompar(void const *, void const *);
static int scompar(void const *, void const *);
static int add_attr(struct dataset *, char const *, char const *);
static void subst(char const *, char *, size_t, char const *, char const *);

static int
conflictp(dh, di, eh, ei)
char const *dh, *eh;
size_t di, ei;
{
	if (dh && eh && !strcmp(dh, eh)) {
		fprintf(stderr, "ci: Duplicated Entry: %s:%s, %s:%s\n", sizetostr(di), dh, sizetostr(ei), eh);
		dwerrno = DWE_DUPHNDL;
		return 1;
	}
	return 0;
}

static int
add_attr(d, attribute, value)
struct dataset *d;
char const *attribute, *value;
{
	size_t i;

	for (i=0; i<d->nattrs; i++) {
		if (!strcasecmp(d->attrs[i].name, attribute)) {
			break;
		}
	}
	if (i==d->nattrs) {
		if (i >= d->maxnattrs) {
			struct attr *new;
			size_t siz = d->maxnattrs + 64;
			if (!(new = realloc(d->attrs, siz * sizeof *d->attrs))) {
				dwerrno = errno;
				return -1;
			}
			d->attrs = new;
			d->maxnattrs = siz;
		}
		if (!(d->attrs[i].name = strdup(attribute))) {
			dwerrno = errno;
			return -1;
		}
		if (!(d->attrs[i].value = strdup(value))) {
			free((void *)d->attrs[i].name);
			dwerrno = errno;
			return -1;
		}
		d->nattrs++;
	}
	else {
		fprintf(stderr, "warning: duplicated attributes [%s] found in corpus [%s]\n", attribute, d->handle);
		dwerrno = DWE_DUPENT;
		return -1;
	}
	return 0;
}

static void
ci_reset()
{
	size_t i;
	struct dataset *d;

	for (i=0; i<ndsets; i++) {
		size_t j;
		d = &datasets[i];
		free((void *)d->handle);
		free((void *)d->shortname);
		free((void *)d->dataroot);
		d->flags = 0;

		assert(d->attrs);
		for (j=0; j<d->nattrs; j++) {
			struct attr *u;
			u = d->attrs+j;
			free((void *)u->name);
			free((void *)u->value);
		}
		free(d->attrs);

		d->nattrs = d->maxnattrs = 0;
		d->attrs = NULL;

		handles[i] = NULL;	/* important! */
		hdatasets[i] = sdatasets[i] = NULL;
	}

	c = NULL;
	ndsets = 0;
}

static char alt_ci_conf[MAXPATHLEN] = "\0";

void
ci_set_ciconf(path)
char const *path;
{
	strlcpy(alt_ci_conf, path, sizeof alt_ci_conf);
}

int
ci_init(projectroot)
char const *projectroot;
{
	size_t i;
	struct dataset *d = NULL;
	char ci_conf[MAXPATHLEN];
	size_t line;
	FILE *f = NULL;
	static char buf[BSIZE];

	if (ndsets) {	/* initialization flag is resetted. */
		fprintf(stderr, "ci: resetting.\n");
		ci_reset();
	}

	if (*alt_ci_conf) {
		strlcpy(ci_conf, alt_ci_conf, sizeof ci_conf);
	}
	else {
		snprintf(ci_conf, sizeof ci_conf, "%s/etc/ci.conf", projectroot);
	}
	if (!(f=fopen(ci_conf, "r"))) {
		nperror(ci_conf);
		dwerrno = errno;
		goto error;
	}
	for (line=0; fgets(buf, sizeof buf, f); line++) {
		char *attribute, *value, *p;

		if (p = index(buf, '\n')) {
			*p = '\0';
		}
		for (p=buf; isspace(*p&0xff); p++) ;
		if (!*p || *p=='#') {
			continue;
		}
		attribute = p;
		if (!(value = index(attribute, ':'))) {
			fprintf(stderr, "%s:%s: `:' expected\n", ci_conf, sizetostr(line));
			dwerrno = DWE_MALFCI;
			goto error;
		}
		*value++ = '\0';
		for (; isspace(*value&0xff); value++) ;
		if (!*value) {
			fprintf(stderr, "%s:%s: value is empty. attribute %s ignored.\n", ci_conf, sizetostr(line), attribute);
			dwerrno = DWE_MALFCI;
			goto error;
		}

		if (!strcasecmp(CI_HANDLE, attribute)) {
			if (!d) {
				d = datasets;
			}
			else {
				d++;
			}
			if (d>=datasets+nelems(datasets)) {
				fprintf(stderr, "too many handles\n");
				dwerrno = DWE_MFILE;
				goto error;
			}

			/* set default values */
			d->handle = NULL;
			d->shortname = NULL;
			d->flags = 0;
			d->nattrs = d->maxnattrs = 0;
			d->attrs = NULL;

			/* it's safe to increment ndsets now. */
			ndsets++;

			if (!(d->handle = strdup(value))) {
				dwerrno = errno;
				goto error;
			}
			if (add_attr(d, attribute, d->handle)==-1) {
				goto error;
			}
		}

		else if (!d) {
			/* ignore */
			fprintf(stderr, "%s:%s: entry ignored: \"%s:%s\"\n", ci_conf, sizetostr(line), attribute, value);
		}

		else {
			if (!strcasecmp(CI_SHORTNAME, attribute)) {
				if (!(d->shortname = strdup(value))) {
					dwerrno = errno;
					goto error;
				}
			}
			else if (!strcasecmp(CI_DATAROOT, attribute)) {
#if 1
				char tmp[MAXPATHLEN], dr[MAXPATHLEN];
#define	V_HANDLE	"${HANDLE}"
#define	V_GETAROOT	"${GETAROOT}"

				subst(value, tmp, sizeof tmp, V_HANDLE, d->handle);
				subst(tmp, dr, sizeof dr, V_GETAROOT, projectroot);
#endif
				if (!(d->dataroot = value = strdup(value))) {
					dwerrno = errno;
					goto error;
				}
			}
			else if (!strcasecmp(CI_FLAGS, attribute)) {
				d->flags = strtoul(value, NULL, 16);
			}
			if (add_attr(d, attribute, value)==-1) {
				goto error;
			}
		}
	}
	fclose(f);
	f=NULL;

	if (!d) {
		fprintf(stderr, "%s is empty\n", ci_conf);
		dwerrno = DWE_EMPTYCI;
		goto error;
	}

	for (i=0, d=datasets; i<ndsets; d++, i++) {
		struct dataset *e;
		size_t di = d-datasets+1;
		for (e=d+1; e<datasets+ndsets; e++) {
			size_t ei = e-datasets+1;
			if (conflictp(d->handle, di, e->handle, ei) ||
			    conflictp(d->handle, di, e->shortname, ei) ||
			    conflictp(d->shortname, di, e->handle, ei) ||
			    conflictp(d->shortname, di, e->shortname, ei)) {
				goto error;
			}
		}
		handles[i] = d->handle;
		hdatasets[i] = sdatasets[i] = d;
		assert(d->nattrs);
		assert(d->attrs);
		if (!d->dataroot) {
			char dataroot[MAXPATHLEN];
			snprintf(dataroot, sizeof dataroot, "%s/data/%s/", projectroot, d->handle);
			if (!(d->dataroot = strdup(dataroot))) {
				dwerrno = errno;
				goto error;
			}
			if (add_attr(d, CI_DATAROOT, d->dataroot)==-1) {
				goto error;
			}
		}
		qsort(d->attrs, d->nattrs, sizeof *d->attrs, ucompar);
	}
	handles[i] = NULL;

	qsort(hdatasets, ndsets, sizeof *hdatasets, hcompar);
	qsort(sdatasets, ndsets, sizeof *sdatasets, scompar);

	return 0;
error:
	ci_reset();
	if (f) {
		fclose(f);
	}
	return -1;
}

static int
ucompar(va, vb)
void const *va, *vb;
{
	struct attr const *a, *b;
	a = va;
	b = vb;
	return strcasecmp(a->name, b->name);
}

static int
hcompar(va, vb)
void const *va, *vb;
{
	struct dataset * const *a, * const *b;
	a = va;
	b = vb;
	return strcmp((*a)->handle, (*b)->handle);
}

static int
scompar(va, vb)
void const *va, *vb;
{
	struct dataset * const *a, * const *b;
	a = va;
	b = vb;
	if (!(*a)->shortname && !(*b)->shortname) {
		return 0;
	}
	if (!(*a)->shortname) {
		return -1;
	}
	if (!(*b)->shortname) {
		return 1;
	}
	return strcmp((*a)->shortname, (*b)->shortname);
}

static int
set_dictionary(handle)
char const *handle;
{
	struct dataset key, *keyp, **cp;

	if (c && !strcmp(handle, c->handle)) { /* already set */
		return 0;
	}

	/* search for new dataset */
	if (!handle) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	key.handle = key.shortname = handle;
	keyp = &key;
	if (!(cp=bsearch(&keyp, hdatasets, ndsets, sizeof keyp, hcompar)) &&
	    !(cp=bsearch(&keyp, sdatasets, ndsets, sizeof keyp, scompar))) {
		fprintf(stderr, "ci:set_dictionary: no such corpus: handle = %s\n", handle);
		/* c is not changed */
		dwerrno = DWE_NOHNDL;
		return -1;
	}
	c = *cp;

	return 0;
}

char const **
ci_handles()
{
	return handles;
}

char const *
ci_value(handle, attribute)
char const *handle, *attribute;
{
	struct attr *u, key;

	if (set_dictionary(handle)==-1) {
		return NULL;
	}
	assert(c->attrs);

	key.name = attribute;

	if (u = bsearch(&key, c->attrs, c->nattrs, sizeof *c->attrs, ucompar)) {
		return u->value;
	}
	dwerrno = DWE_NOENT;
	return NULL;
}

u_int
ci_flags(handle)
char const *handle;
{
	if (set_dictionary(handle)==-1) {
		return 0;
	}
	return c->flags;
}

void
ci_print_dataset(stream, handle)
FILE *stream;
char const *handle;
{
	size_t i;

	if (set_dictionary(handle)==-1) {
		return;
	}

	fprintf(stream, "handle: %s\n", c->handle);
	fprintf(stream, "shortname: %s\n", c->shortname?c->shortname:"(null)");
	fprintf(stream, "flags: 0x%x\n", c->flags);

	assert(c->attrs);
	for (i=0; i<c->nattrs; i++) {
		struct attr *u;
		u = c->attrs+i;
		if (!strcmp(u->name, CI_HANDLE) || !strcmp(u->name, CI_SHORTNAME)) {
			continue;
		}
		fprintf(stream, "%s: %s\n", u->name, u->value);
	}
}

static void
subst(s, t, n, old, new)
char const *s, *old, *new;
char *t;
size_t n;
{
	size_t lo, ln;
	size_t l;
	char const *p, *q;

	lo = strlen(old);
	ln = strlen(new);

	if (n==0) {
		return;
	}
	strcpy(t, "");
	for (p = s; q=strstr(p, old); p = q) {
		l = q-p;
		if (l<n-1) memmove(t, p, l), t += l;
		l = ln;
		if (l<n-1) memmove(t, new, l), t += l;
		q += lo;
	}
	l = strlen(p);
	if (l<n-1) memmove(t, p, l), t += l;
	*t = '\0';
}

#endif	/* USE_CICONF */
