/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dcw.c,v 1.6 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/dcw.h>

#include "dwsys.h"
#include "dfi.h"
#include "dcwP.h"
#include "sdup.h"
#include "nht.h"
#include "nsxhash.h"

struct dw_FILE {
	int magic;

	void *p;
	size_t len;

	struct dcw_hdr h, *hp;
	struct bindex_t *b;
	struct rindex_t *r;
	char *fullwd;

/*
	char path[MAXPATHLEN];
*/
	char const *handle;

	struct {
		struct dfilt *s;
	} filt;
};

struct rcompar_key {
	char const *fullwd;
	char const *name;
};

static int rcompar(void const *, void const *);

static int check_dcw_hdr(struct dcw_hdr *);
static void const *h_lookup(struct rcompar_key const *, void const *, size_t);

DWFILE *
dcw_open(path, handle)
char const *path, *handle;
{
	int d = -1;
	DWFILE *f = NULL;
	struct stat sb;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}

	f->p = MAP_FAILED;
	f->len = 0;
/*
	strlcpy(f->path, path, sizeof f->path);
*/
	f->handle = handle;

	if ((d = open(path, O_RDONLY))==-1) {
		dwerrno = errno;
		goto err;
	}

	if (fstat(d, &sb)==-1) {
		dwerrno = errno;
		goto err;
	}

	f->len = sb.st_size;
/* AAA */
	if ((f->p = mmap(0, f->len, PROT_READ, MAP_SHARED, d, 0))==MAP_FAILED) {
		dwerrno = errno;
		goto err;
	}

	close(d);
	d = -1;

	f->hp = f->p;
	memmove(&f->h, f->hp, sizeof f->h);
	f->b = (struct bindex_t *)((char *)f->p+f->h.bindex_offset);
	f->r = (struct rindex_t *)((char *)f->p+f->h.rindex_offset);
	f->fullwd = (char *)f->p + f->h.fullwd_offset;

	if (check_dcw_hdr(&f->h)==-1) {
		goto err;
	}

	f->filt.s = NULL;

	f->magic = DCW_MAGIC;
	return f;
err:
	if (d!=-1) {
		close(d);
	}
	dcw_close(f);
	return NULL;
}

static int
check_dcw_hdr(h)
struct dcw_hdr *h;
{
	switch (h->type) {
	case DCW_ROW:
		if (strcmp(h->ident, "!dcw.row")) {
			dwerrno = DWE_BADF;
			return -1;
		}
		break;
	case DCW_COL:
		if (strcmp(h->ident, "!dcw.col")) {
			dwerrno = DWE_BADF;
			return -1;
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (h->version != DCW_VERSION) {
		dwerrno = DWE_BADVER;
		return -1;
	}
	return 0;
}

int
dcw_close(f)
DWFILE *f;
{
	int e = 0;
	f->magic = 0;
	if (f->p!=MAP_FAILED) {
		e = munmap(f->p, f->len);
	}
	free(f);
	return e;
}

int
dcw_mtime(f)
DWFILE *f;
{
	return f->h.mtime;
}

ssize_t
dcw_size(f)
DWFILE *f;
{
	return f->h.nentries;
}

char const *
dcw_id2name(f, id)
DWFILE *f;
idx_t id;
{
	struct dfilt *fil;
	if (id==0 || id>f->h.nentries) {
		dwerrno = DWE_BADR;
		return NULL;
	}
	if ((fil = f->filt.s) && (*fil->screened)(fil->cookie, id)) {
		dwerrno = DWE_BADR;
		return NULL;
	}

	return f->fullwd+f->b[id].b;
}

idx_t
dcw_name2id(f, name)
DWFILE *f;
char const *name;
{
	struct rcompar_key key;
	struct rindex_t const *p;
	idx_t id;
	struct dfilt *fil;

	key.name = name;
	key.fullwd = f->fullwd;

	switch (f->h.rindex_type) {
	case RIT_BSEARCH:
		if (!(p = bsearch(&key, f->r+1, f->h.nentries, sizeof *f->r, rcompar))) {
			dwerrno = DWE_NOENT;
			return 0;
		}
		break;
	case RIT_HASH:
		if (!(p = h_lookup(&key, f->r, f->h.size))) {
			return 0;
		}
		break;
	default:
		dwerrno = DWE_BADF;
		return 0;
	}
	id = p->r;
	if ((fil = f->filt.s) && (*fil->screened)(fil->cookie, id)) {
		dwerrno = DWE_BADR;
		return 0;
	}
	return id;
}

static int
rcompar(va, vb)
void const *va, *vb;
{
	struct rcompar_key const *a;
	struct rindex_t const *b;

	a = va;
	b = vb;
	return strcmp(a->name, a->fullwd+b->b);
}

static void const *
h_lookup(key, base, size)
struct rcompar_key const *key;
void const *base;
size_t size;
{
	size_t s, c;
	char const *p;
	struct rindex_t const *b, *base1;

	p = key->name;
	base1 = base;
	c = 0;
	for (s = nsxhash(size, p, c++); ; s = nsxhash(size, p, c++)) {
		b = &base1[s];
		if (!b->b) {	/* end of chain */
			break;	/* not found */
		}
		else if (!strcmp(p, key->fullwd+b->b)) {	/* found */
			return b;
		}
		/* follow chain */
	}
	dwerrno = DWE_NOENT;
	return NULL;
}

int
dcw_setf(f, s)
DWFILE *f;
struct dfilt *s;
{
	f->filt.s = s;
	return 0;
}

char const *
dcw_handle(f)
DWFILE *f;
{
	return f->handle;
}
