/*
 *	Copyright (c) 2005 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: dfs.c,v 1.7 2007/03/29 07:37:38 nis Exp $";
#endif

#include <config.h>

#include <sys/param.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>

#include "nnet.h"
#include "dfs.h"

struct sfile {
	FILE *f;		/* fopen */
	int s;			/* dfs */
	char *base, *bp, *ep;	/* MMAP */

	size_t size, recv;
	size_t pos;
	char buf[8192];
	char *p, *e;
};

char const *dfs_host = NULL;
char const *dfs_proto = NULL;

static ssize_t getfs(int, char const *);

#if ! defined(MAP_FILE)
#define MAP_FILE	0
#endif

#if ! defined(MIN)
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

SFILE *
sfopen(path, mode)
char const *path;
char const *mode;
{
	SFILE *f;
	if (strcmp(mode, "r")) {
		return NULL;
	}
	if (!(f = calloc(1, sizeof *f))) {
		return NULL;
	}
	f->f = NULL;
	f->s = -1;
	f->base = MAP_FAILED;
	f->bp = f->ep = NULL;
	f->pos = 0;
	f->recv = 0;
	f->p = f->e = f->buf;
	if (*path != '!') {
#if 0
		int d;
		struct stat sb;
		if ((d = open(path, O_RDONLY)) == -1) {
			perror(path);
			return NULL;
		}
		if (fstat(d, &sb) == -1) {
			perror(path);
			return NULL;
		}
		f->size = sb.st_size;
		if ((f->base = mmap(0, f->size, PROT_READ, MAP_FILE, d, 0)) == MAP_FAILED) {
			perror(path);
			return NULL;
		}
		f->bp = f->base;
		f->ep = f->base + f->size;
		madvise(f->base, f->size, MADV_SEQUENTIAL);
		close(d);
		f->p = f->bp;
		f->e = f->ep;
#else
		if (!(f->f = fopen(path, mode))) {
			free(f);
			return NULL;
		}
		fseeko(f->f, 0, SEEK_END);
		f->size = ftello(f->f);
		rewind(f->f);
#endif
	}
	else {
		static int s0 = -1;
		if (s0 == -1) {
			if (!dfs_host) {
				fprintf(stderr, "dfs_host is not set\n");
				return NULL;
			}
			if (!dfs_proto) {
				fprintf(stderr, "dfs_proto is not set\n");
				return NULL;
			}
			if ((s0 = nnet_cli(dfs_host, dfs_proto)) == -1) {
				free(f);
				return NULL;
			}
		}
		f->s = s0;
		if ((f->size = getfs(f->s, path + 1)) == -1) {
			free(f);
			return NULL;
		}
	}
	return f;
}

/*
 * return value:
 * 0: success
 * -1: fail
 * 1: unexpected eof
 */
int
sfclose(f)
SFILE *f;
{
	int e;
	if (f->base != MAP_FAILED) {
		madvise(f->base, f->size, MADV_DONTNEED);
		e = munmap(f->base, f->size);
	}
	else if (f->f) {
		e = fclose(f->f);
	}
	else if (f->s != -1) {
		int i;
/* fprintf(stderr, "flushing... "); */
		e = 0;
		for (i = 0; f->pos != f->size; ) {
/* fprintf(stderr, "%d", i); */
			if (sffillbuf(f) == 0) {
				e = 1;
				break;
			}
		}
/* fprintf(stderr, "done"); */
	}
	else {
		fprintf(stderr, "sfclose: internal error\n");
		e = -1;
	}
	free(f);
	return e;
}

char *
sfgets(buf, size, f)
char *buf;
size_t size;
SFILE *f;
{
	size_t i;
	if (f->pos == f->size) {
		return NULL;
	}
	for (i = 0; i < size - 1; i++) {
		if (f->p == f->e && sffillbuf(f) == 0) {
			break;
		}
		buf[i] = *f->p++;
		if ((buf[i]&0xff) == '\n') {
			i++;
			break;
		}
	}
	buf[i] = '\0';
	f->pos += i;
	return buf;
}

int
sfpeekc(f)
SFILE *f;
{
	if (f->pos == f->size) {
		/* normal eof */
		return EOF;
	}
	if (f->p == f->e && sffillbuf(f) == 0) {
		/* emergency! */
		return EOF;
	}
	return *f->p & 0xff;
}

size_t
sfpos(f)
SFILE *f;
{
	return f->pos;
}

size_t
sfsize(f)
SFILE *f;
{
	return f->size;
}

size_t
sffillbuf(f)
SFILE *f;
{
	ssize_t l;

	if (f->base != MAP_FAILED) {
		return 0;
	}
	l = MIN(f->size - f->recv, sizeof f->buf);
/* fprintf(stderr, "sffillbuf: rest = %d, l = %d\n", (int)(f->size - f->recv), (int)l); */
	if (l == 0) {
		return 0;
	}
/*
	if (f->base != MAP_FAILED) {
		assert(f->bp + l <= f->ep);
		memcpy(f->buf, f->bp, l);
		f->bp += l;
	}
	else */
	if (f->f) {
		l = fread(f->buf, 1, l, f->f);
	}
	else if (f->s != -1) {
		l = recv(f->s, f->buf, l, MSG_WAITALL);
	}
	else {
		fprintf(stderr, "sfillbuf: internal error\n");
		l = 0;
	}
	if (l <= 0) {
		return 0;
	}

	f->p = f->buf;
	f->e = f->buf + l;
	f->recv += l;
	return l;
}

static ssize_t
getfs(s, name)
char const *name;
{
	char path[FSIZE];
	long long len;	/* XXX must be fixed */

	if (name) {
		strlcpy(path, name, sizeof path);
	}
	else {
		bzero(path, sizeof path);
	}
	if (send(s, path, sizeof path, 0) != sizeof path) {
		return -1;
	}
	if (!name) {
		return 0;
	}

	if (recv(s, &len, sizeof len, MSG_WAITALL) != sizeof len) {
		return -1;
	}

	return len;
}
