/*
 *	Copyright (C) 1997 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dumpwam.c,v 1.2 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

static void usage(void);

static void print_summary(FILE *, WAM *);

static void dump_cw(FILE *, WAM *, idx_t, int);
static void eent(FILE *, WAM *, idx_t, int);

static void dump_xr(FILE *, WAM *, idx_t, int);
static void edump(FILE *, WAM *, idx_t, int);

char *progname;

int header_only = 0;
int number = 0;
int no_key_translation = 0;

static void
usage()
{
	fprintf(stderr, "usage: %s [-n] [-h] [-k] handle summary|cw_row|cw_col|xr_row|xr_col [id]\n", progname);
	exit(1);
}

main(argc, argv)
char *argv[];
{
	int dir;
	FILE *stream = stdout;
	char *handle;
	idx_t id = 0;
	int type = 0;
	WAM *w;

	wam_init(NULL);

	progname = basename(argv[0]);

	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(argv[0], "-n")) number = 1;
		else if (!strcmp(argv[0], "-h")) header_only = 1;
		else if (!strcmp(argv[0], "-k")) no_key_translation = 1;
		else break;
	}

	if (argc!=2 && argc!=3) {
		usage();
	}

	handle = argv[0];

	if (!strcmp(argv[1], "summary")) {
		type = 0;
		/* argv[2] will be silently ignored */
	}
	else if (!strcmp(argv[1], "cw_row")) {
		type = 1;
	}
	else if (!strcmp(argv[1], "cw_col")) {
		type = 2;
	}
	else if (!strcmp(argv[1], "xr_row")) {
		type = 3;
	}
	else if (!strcmp(argv[1], "xr_col")) {
		type = 4;
	}
	else {
		usage();
	}

	if (argc==3) {
		id = strtoull(argv[2], NULL, 10);
	}

	if (!(w = wam_open(handle))) {
		exit(1);
	}

	if (type==0 || type==1 || type==3) {
		dir = WAM_ROW;
	}
	else {
		dir = WAM_COL;
	}

	if (type==0) {
		print_summary(stream, w);
	}
	else if (type==1 || type==2) {
		dump_cw(stream, w, id, dir);
	}
	else {
		dump_xr(stream, w, id, dir);
	}

	wam_close(w);
	return 0;
}

static void
print_summary(stream, w)
FILE *stream;
WAM *w;
{

#define	foo0(name, dir) \
	fprintf(stream, #name ": %d\n", name(w, dir));

#define	bar(dir)	\
	fprintf(stream, "*** " #dir " ***\n"); \
	foo0(wam_size, WAM_##dir); \
	foo0(wam_max_elem_num, WAM_##dir); \
	foo0(wam_max_freq_sum, WAM_##dir); \
	foo0(wam_total_elem_num, WAM_##dir); \
	foo0(wam_total_freq_sum, WAM_##dir); \
/*
	fprintf(stream, "--- CW_" #dir " ---\n"); \
	cw_print_info(WAM_XR(w, WAM_##dir), stream); \
	fprintf(stream, "--- XR_" #dir " ---\n"); \
	xr_print_info(WAM_XR(w, WAM_##dir), stream, 0); \
*/

	bar(ROW);
	bar(COL);

}

static void
dump_cw(stream, w, id, dir)
FILE *stream;
WAM *w;
{
	size_t nrsyms;

	if (id) {
		number = 0;
		eent(stream, w, id, dir);
	}
	else {
		nrsyms = wam_size(w, dir);

		fprintf(stream, "nrsyms = %d\n", nrsyms);
		for (id=1; id<=nrsyms; id++) {
			eent(stream, w, id, dir);
		}
	}
}

static void
eent(stream, w, id, dir)
FILE *stream;
WAM *w;
idx_t id;
{
	if (number) fprintf(stream, "%d ", id);
	fprintf(stream, "%s\n", wam_id2name(w, dir, id));
}

static void
dump_xr(stream, w, id, dir)
FILE *stream;
WAM *w;
{
	size_t n;

	if (id==0) {
		idx_t seq;
		n = wam_size(w, dir);
		for (seq=1; seq<=n; seq++) {
			edump(stream, w, seq, dir);
		}
	}
	else {
		edump(stream, w, id, dir);
	}
}

static void
edump(stream, w, id, dir)
FILE *stream;
WAM *w;
idx_t id;
{
	freq_t f;
	char const *key;
	struct xr_vec const *vec=NULL;
	freq_t elem_num;
	freq_t freq_sum;
	size_t r;
	freq_t n;
	int rdir = WAM_REVERT_DIRECTION(dir);

	if (no_key_translation) {
		key = "*";
	}
	else {
		if (!(key = wam_id2name(w, dir, id))) {
			key = "(NULL)";
		}
	}

	if ((elem_num = wam_get_vec(w, dir, id, &vec))==-1) {
		exit(1);
	}
	fprintf(stream, "entry %d(%s): elem_num=%d, freq_sum=%ld\n", id, key, elem_num, vec->freq_sum);

	if (header_only) {
		return;
	}

/* consistency check */
	freq_sum = vec->freq_sum;
	n = vec->elem_num;
	f = wam_freq_sum(w, dir, id);
	if (f!=-1 && f!=freq_sum) {
		fprintf(stderr, "dumpxr: internal error(f(%d)!=freq_sum(%d))\n", f, freq_sum);
		exit(1);
	}
	if (n!=elem_num) {
		fprintf(stream, "dumpxr: warning: vec->elem_num(%d)!=elem_num(%d)\n", n, elem_num);
	}

/* dump elems */
	for (r=0; r<elem_num; r++) {
		struct xr_elem const *elem;
		idx_t ii;
		freq_t freq;
		elem = &vec->elems[r];
		ii = elem->id;
		freq = elem->freq;
		if (ii==0) {
			key = "NULL";
		}
		else if (no_key_translation) {
			key = "*";
		}
		else {
			key = wam_id2name(w, rdir, ii);
		}
		fprintf(stream, "%d/%d: %d(%s), %d\n", id, r, ii, key, freq);
	}
}
