/*
 *	Copyright (C) 1999 Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwam.c,v 1.17 2007/03/29 08:34:55 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <assert.h>
#include <netdb.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <time.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

#include "dxrP.h"
#include "tutil.h"
#include "butil.h"
#include "gstat.h"

#if HAVE_IEEEFP_H
#include <ieeefp.h>
#endif

struct wgv_cache;

struct wam_d {
	char handle[MAXPATHLEN];
	struct dwsched ds;
	DWFILE *cw[2], *xr[2];
	DWFILE *a;
	struct {
		ssize_t summary_interval;
		int cascade, prob_mode, fault_tolerant;
	} param;
	struct {
		struct sym_vec *sv;
		ssize_t np;
	} s[2];
	struct xr_vec *v[2];
	ssize_t vsize[2];
	struct wgv_cache *cache;
};

static char proot[MAXPATHLEN];

static void reset_sv(WAM *);
static int sum_conn(WAM *, int, struct rcxsp **);
static ssize_t rdgetvec(WAM *, struct rcxsp *, size_t, idx_t, struct xr_vec *, ssize_t);
static int ecompar_static(void const *, void const *);

static int rdsetmask(WAM *, struct rcxsp *, size_t, idx_t *, size_t, idx_t *, size_t);

static struct wgv_cache *wgv_cache_init(int);
static int wgv_cache_finish(struct wgv_cache *);
static int wgv_cache_insert(struct wgv_cache *, int, idx_t, struct xr_vec const *);
static ssize_t wgv_cache_lookup(struct wgv_cache *, int, idx_t, struct xr_vec const **);
#if defined GAM_TEST
static void report_query(struct syminfo const *, size_t, WAM *, int, int);
#endif

int
wam_init(projectroot)
char const *projectroot;
{
#if 0 && HAVE_IEEEFP_H && defined fpsetprec
	fpsetprec(FP_PE);
#endif
	if (!projectroot && !(projectroot = getenv(PROOTENV))) {
		dwerrno = DWE_NOROOT;
		return -1;
	}
#if USE_CICONF
	if (ci_init(projectroot)==-1) {
		return -1;
	}
#endif
	if (wsh_init()==-1) {
		return -1;
	}
#ifdef GSTAT
	gstat_init();
#endif
	strlcpy(proot, projectroot, sizeof proot);
	return 0;
}

WAM *
wam_open(handle)
char const *handle;
{
	WAM *f;

	if (index(handle, '/')) {	/* XXX: utf-8 de multi-byte ni gisou sareruto koredemo kiken? */
		return NULL;
	}

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}
	f->cw[WAM_ROW] = f->cw[WAM_COL] = NULL;
	f->xr[WAM_ROW] = f->xr[WAM_COL] = NULL;
	f->a = NULL;
	f->s[0].sv = f->s[1].sv = NULL;
	f->v[0] = f->v[1] = NULL;
	f->vsize[0] = f->vsize[1] = 0;
	f->cache = NULL;

	strlcpy(f->handle, handle, sizeof f->handle);

	if (read_dwsched(&f->ds, proot, f->handle)==-1) {
		goto error;
	}

	if (f->ds.local) {
		if (f->ds.rcx[0].np!=1 || f->ds.rcx[1].np!=1) {
			dwerrno = DWE_INVAL;
			goto error;
		}
	}
	else {
		f->param.prob_mode = 0;
		f->param.cascade = 0;
		f->param.summary_interval = 0;
		f->param.fault_tolerant = 0;
	}

	return f;
error:

	if (f) {
		wam_close(f);
	}

	return NULL;
}

static void
reset_sv(f)
WAM *f;
{
	int i;
	for (i=0; i<2; i++) {
		free(f->s[i].sv);
		f->s[i].sv = NULL;
	}
}

void
wam_close(f)
WAM *f;
{
	if (f->cw[WAM_ROW]) {
		(void)dcw_close(f->cw[WAM_ROW]);
	}
	if (f->cw[WAM_COL]) {
		(void)dcw_close(f->cw[WAM_COL]);
	}
	if (f->xr[WAM_ROW]) {
		(void)dxr_close(f->xr[WAM_ROW]);
	}
	if (f->xr[WAM_COL]) {
		(void)dxr_close(f->xr[WAM_COL]);
	}
	if (f->a) {
		(void)dau_close(f->a);
	}
	if (!f->ds.local) {
		reset_sv(f);
	}
	if (f->cache) {
		(void)wgv_cache_finish(f->cache);
		f->cache = NULL;
	}
	clear_dwsched(&f->ds);
	bzero(f, sizeof *f);
	free(f);
}

#define	CWR(f, d) (f->cw[d] || (f->cw[d] = dcw_open(f->ds.rcx[d].cw_path, f->handle)))

/* use non-cascade rems */
#define	XRR(f, d) (f->xr[d] || (f->xr[d] = dxr_open(f->ds.rcx[d].rems[0][0].host, f->handle)))

ssize_t
wam_size(w, dir)
WAM *w;
{
	if (!CWR(w, dir)) {
		return -1;
	}
	return dcw_size(w->cw[dir]);
}

char const *
wam_id2name(w, dir, id)
WAM *w;
idx_t id;
{
	if (!CWR(w, dir)) {
		return NULL;
	}
	return dcw_id2name(w->cw[dir], id);
}

idx_t
wam_name2id(w, dir, name)
WAM *w;
char const *name;
{
	if (!CWR(w, dir)) {
		return 0;
	}
	return dcw_name2id(w->cw[dir], name);
}

ssize_t
wam_elem_num(w, dir, id)
WAM *w;
idx_t id;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_elem_num(w->xr[dir], id);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_elem_num(sp->f, id);
	}
}

freq_t
wam_freq_sum(w, dir, id)
WAM *w;
idx_t id;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_freq_sum(w->xr[dir], id);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_freq_sum(sp->f, id);
	}
}

ssize_t
wam_max_elem_num(w, dir)
WAM *w;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_max_elem_num(w->xr[dir]);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_max_elem_num(sp->f);
	}
}

freq_t
wam_max_freq_sum(w, dir)
WAM *w;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_max_freq_sum(w->xr[dir]);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_max_freq_sum(sp->f);
	}
}

freq_t
wam_total_elem_num(w, dir)
WAM *w;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_total_elem_num(w->xr[dir]);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_total_elem_num(sp->f);
	}
}

freq_t
wam_total_freq_sum(w, dir)
WAM *w;
{
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return -1;
		}
		return dxr_total_freq_sum(w->xr[dir]);
	}
	else {
		struct rcxsp *sp;
		if (sum_conn(w, dir, &sp)==-1) {
			return -1;
		}
		return dwc_dxr_total_freq_sum(sp->f);
	}
}

static int
sum_conn(w, dir, spp)
WAM *w;
struct rcxsp **spp;
{
	struct rcxs *rp;
	struct rcxsp *sp;
	rp = &w->ds.rcx[dir];
	sp = &rp->s;

	if (!sp->f && rcxsp_connect_n_open(sp, w->param.fault_tolerant)==-1) {
		if (w->param.fault_tolerant) {
			ssize_t i;
syslog(LOG_DEBUG, "%s:%s is down.", sp->host, sp->serv);
			for (i = 0; i < rp->np; i++) {
				struct rcxsp *altsp = &rp->rems[0][i];
				strlcpy(sp->host, altsp->host, sizeof sp->host);
				strlcpy(sp->serv, altsp->serv, sizeof sp->serv);
syslog(LOG_DEBUG, "trying alternte node: %s:%s...", sp->host, sp->serv);
				if (rcxsp_connect_n_open(sp, w->param.fault_tolerant) != -1) {
syslog(LOG_DEBUG, "success.");
					goto success;
				}
			}
		}
		return -1;
	}
success:
	*spp = sp;
	return 0;
}

struct syminfo *
wsh(q, nq, w, dir, weight_type, nd, total)
struct syminfo const *q;
ssize_t nq;
WAM *w;
ssize_t *nd;
ssize_t *total;
{
#if defined GAM_TEST
	report_query(q, nq, w, dir, weight_type);
#endif
	struct syminfo *d;
	if (w->ds.local) {
		int rdir = WAM_REVERT_DIRECTION(dir);
		if (!XRR(w, dir)) {
			return NULL;
		}
		if (!XRR(w, rdir)) {
			return NULL;
		}
		if (!(d = dw_wsh(q, nq, w->xr[dir], w->xr[dir], w->xr[rdir], weight_type, nd, total))) {
			return NULL;
		}
	}
	else {
		ssize_t summary_interval;
		int prob_mode;
		struct rcxs *rp;
		int croot;

		summary_interval = w->param.summary_interval;
		prob_mode = w->param.prob_mode;

		rp = &w->ds.rcx[dir];

		croot = w->param.cascade ? -1 : -2;

		if (!w->s[dir].sv && (w->s[dir].np = makesv(croot, rp->np, rp->rems, &w->s[dir].sv, NULL, w->param.fault_tolerant))==-1) {
			return NULL;
		}
		if (!(d = rdwsh(q, nq, w->s[dir].sv, w->s[dir].np, weight_type, nd, *nd, total, summary_interval, prob_mode, w->param.fault_tolerant))) {
			return NULL;
		}
	}

	qsort(d, (*nd), sizeof *d, sym_compar);

	return d;
}

struct syminfo *
bex_wsh(q, nq, w, dir, weight_type, nd, total, e, ne)
ssize_t nq, ne;
struct syminfo const *q;
struct syminfo const *e;
WAM *w;
ssize_t *nd, *total;
{
	struct syminfo *o;
	struct syminfo *v;
	ssize_t no;

	if (!(o = bex_conv(q, nq, e, ne, &no))) {
		return NULL;
	}
	v = wsh(o, no, w, dir, weight_type, nd, total);
	free(o);
	return v;
}

char dl_path[MAXPATHLEN];

int
wam_setopt(w, optname, optval)
WAM *w;
char *optval;
{
	int n;
	switch (optname) {
	case WAM_OFAULTTOLERANT:
		snprintf(dl_path, sizeof dl_path, "%s/etc/donwlist.lst", proot);
		if (w->ds.local) {
			dwerrno = DWE_INVAL;
			return -1;
		}
		n = *(int *)optval;
		if (w->param.fault_tolerant != n) {
			reset_sv(w);
			w->param.fault_tolerant = n;
		}
		w->param.cascade = 0;
		break;
	case WAM_OCASCADE:
		if (w->ds.local || w->param.fault_tolerant) {
			dwerrno = DWE_INVAL;
			return -1;
		}
		n = *(int *)optval;
		if (w->param.cascade != n) {
			reset_sv(w);
			w->param.cascade = n;
		}
		break;
	case WAM_OSUMI:
		if (w->ds.local) {
			dwerrno = DWE_INVAL;
			return -1;
		}
		w->param.summary_interval = *(int *)optval;
		break;
	case WAM_OPROBM:
		if (w->ds.local) {
			dwerrno = DWE_INVAL;
			return -1;
		}
		w->param.prob_mode = *(int *)optval;
		break;
	case WAM_OCACHE:
		if (w->cache) {
			(void)wgv_cache_finish(w->cache);
			w->cache = NULL;
		}
		n = *(int *)optval;
		if (n>0 && !(w->cache = wgv_cache_init(n))) {
			dwerrno = DWE_INVAL;
			return -1;
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
	return 0;
}

WAM *
wam_prop_open(handle, dir, abase)
char const *handle, *abase;
{
	WAM *f;

	if (!(f = wam_open(handle))) {
		return NULL;
	}

	if (!(f->a = dau_open(proot, handle, abase))) {
		wam_close(f);
		return NULL;
	}
	return f;
}

WAM *
wam_prop_cmd(handle, dir, cmd)
char const *handle, *cmd;
{
	WAM *f;

	if (!(f = wam_open(handle))) {
		return NULL;
	}

	if (!CWR(f, dir)) {
		wam_close(f);
		return NULL;
	}

	if (!(f->a = dau_open_cmd(handle, cmd, f->cw[dir]))) {
		wam_close(f);
		return NULL;
	}
	return f;
}

WAM *
wam_prop_dl(handle, dir, path, name, mode)
char const *handle, *path, *name;
{
	WAM *f;

	if (!(f = wam_open(handle))) {
		return NULL;
	}

	if (!CWR(f, dir)) {
		wam_close(f);
		return NULL;
	}

	if (!(f->a = dau_open_dl(handle, path, name, f->cw[dir], mode))) {
		wam_close(f);
		return NULL;
	}
	return f;
}

ssize_t
wam_prop_gets(f, id, p)
WAM *f;
idx_t id;
char const **p;
{
	return dau_gets(f->a, id, p);
}

ssize_t
wam_prop_nentries(f)
WAM *f;
{
	return dau_nentries(f->a);
}

#if 0
#define DEBUG
#endif

ssize_t
wam_get_vec(w, dir, id, v)
WAM *w;
idx_t id;
struct xr_vec const **v;
{
	ssize_t e;
#if defined DEBUG
struct xr_vec const *v0 = NULL;
fprintf(stderr, "wam_get_vec: A\n");
fprintf(stderr, "w: %p\n", w);
fprintf(stderr, "w->cache: %p\n",w->cache);
#endif

	if (w->cache && (e=wgv_cache_lookup(w->cache, dir, id, v)) != -1) {
#if defined DEBUG
v0 = *v;
#else
		/* do nothing */
#endif
	}
#if ! defined DEBUG
	else
#endif
	if (w->ds.local) {
#if defined DEBUG
fprintf(stderr, "wam_get_vec: B1\n");
#endif
		if (!XRR(w, dir)) {
#if defined DEBUG
fprintf(stderr, "!XRR");
#endif
			return -1;
		}
#if defined DEBUG
fprintf(stderr, "wam_get_vec: B1.1\n");
#endif
		e = dxr_get_vec(w->xr[dir], id, v);
#if defined DEBUG
fprintf(stderr, "wam_get_vec: B1.2\n");
#endif
#if defined DEBUG
if (e==-1) {
	fprintf(stderr, "!DXRGV\n");
}
#endif
	}
	else {
		struct rcxsp *rp;
#if defined DEBUG
fprintf(stderr, "wam_get_vec: B2\n");
#endif
		if (!w->v[dir]) {
			ssize_t n;
			if ((n = wam_max_elem_num(w, dir))==-1) {
#if defined DEBUG
fprintf(stderr, "!MAXE");
#endif
				return -1;
			}
			if (!(w->v[dir] = malloc(sizeof_dxr_vec(n)))) {
				dwerrno = errno;
#if defined DEBUG
fprintf(stderr, "!MALLOC");
#endif
				return -1;
			}
			w->vsize[dir] = n;
		}
#if defined DEBUG
fprintf(stderr, "wam_get_vec: C, %p\n", w->v[dir]);
#endif
		rp = w->ds.rcx[dir].rems[0];	/* use non-cascade rems */
		if ((e = rdgetvec(w, rp, w->ds.rcx[dir].np, id, w->v[dir], w->vsize[dir]))==-1) {
#if 0
fprintf(stderr, "!RDGETV");
#endif
			return -1;
		}
		*v = w->v[dir];
#if defined DEBUG
fprintf(stderr, "wam_get_vec: D: %p, %p\n", w, *v);
#endif
		if (
#if defined DEBUG
!v0 &&
#endif
		  w->cache && wgv_cache_insert(w->cache, dir, id, *v)==-1) {
#if defined DEBUG
fprintf(stderr, "!INS");
#endif
			return -1;
		}
	}
#if defined DEBUG
fprintf(stderr, "wam_get_vec: E\n");
if (v0) {
	if (v0->elem_num != (*v)->elem_num) {
		fprintf(stderr, "wam_get_vec: elem_num mismatch\n");
	}
	else if (v0->freq_sum != (*v)->freq_sum) {
		fprintf(stderr, "wam_get_vec: freq_sum mismatch\n");
	}
	else {
		idx_t i;
		for (i = 0; i<v0->elem_num; i++) {
			if (v0->elems[i].id != (*v)->elems[i].id) {
				fprintf(stderr, "wam_get_vec: id mismatch\n");
				break;
			}
			if (v0->elems[i].freq != (*v)->elems[i].freq) {
				fprintf(stderr, "wam_get_vec: freq mismatch\n");
				break;
			}
		}
	}
}
fprintf(stderr, "wam_get_vec: G e=%d\n", e);
#endif
	return e;
}

static int
ecompar_static(va, vb)
void const *va, *vb;
{
	struct xr_elem const *a, *b;
	a = va;
	b = vb;
	if (a->id < b->id) {
		return -1;
	}
	else if (a->id > b->id) {
		return 1;
	}
	return 0;
}

/*
 * この関数はいわゆる cat|sort 方式で実装されている.
 * (merge.c:qmerge と同じ方式) これが merge.c:hmerge より速いかどうかは
 * 状況によって変化するので (どうしてもスピードにこだわのるなら) 実測して
 * 速い方を使って頂きたい.
 */
static ssize_t
rdgetvec(w, r, np, id, v, ve)
WAM *w;
struct rcxsp *r;
size_t np;
ssize_t ve;
idx_t id;
struct xr_vec *v;
{
	size_t p;
	ssize_t en;
	freq_t fs;

	en = 0;
	fs = 0;
	for (p=0; p<np; p++) {
		struct xr_vec const *vv;
		struct rcxsp *rp = &r[p];
		ssize_t e;

		if (!rp->f && rcxsp_connect_n_open(rp, w->param.fault_tolerant)==-1) {
			if (w->param.fault_tolerant) {
syslog(LOG_DEBUG, "%s:%s is down. skipped.", rp->host, rp->serv);
				continue;
			}
			return -1;
		}

		if ((e = dwc_dxr_get_vec(rp->f, id, &vv))==-1) {
#if 0
			fprintf(stderr, __FILN__ ": dwc_dxr_get_vec failed");
#endif
			return -1;
		}

		if (en+e > ve) {
			dwerrno = DWE_IERROR;
			return -1;
		}
		memmove(&v->elems[en], &vv->elems[0], e * sizeof vv->elems[0]);
		en += e;
		fs += vv->freq_sum;
	}
	v->elem_num = en;
	v->freq_sum = fs;
	qsort(v->elems, v->elem_num, sizeof *v->elems, ecompar_static);
	return en;
}

char const *
wam_handle(w)
WAM *w;
{
	return w->handle;
}

/* *** */

struct key {
	int dir;
	idx_t id;
};

struct cell {
	struct key key;
	struct {
		struct {
			struct cell *prev, *next;
		} hash;
		struct {
			struct cell *prev, *next;
		} lru;
	} link;
	struct {
		struct xr_vec *v;
		ssize_t size;
	} v;
};

struct wgv_cache {
	ssize_t size;
	ssize_t hash_size;
	struct cell *pool;
	struct cell *hash;
	struct cell mru, *lru;
};

static void setl(struct cell *, struct cell *, struct cell *, struct cell *, struct cell *);
static size_t hash(struct wgv_cache *, struct key *);

static void
setl(p, lp, ln, hp, hn)
struct cell *p, *lp, *ln, *hp, *hn;
{
	p->link.lru.prev = lp;
	p->link.lru.next = ln;
	p->link.hash.prev = hp;
	p->link.hash.next = hn;
}

static struct wgv_cache *
wgv_cache_init(n)
{
	struct cell *p;
	size_t i;
	struct wgv_cache *c;
	assert(n>0);
	if (!(c = calloc(1, sizeof *c))) {
		return NULL;
	}
	c->size = n;
	if (!(c->pool = calloc(c->size, sizeof *c->pool))) {
		free(c);
		return NULL;
	}
	c->hash_size = c->size*2;
	if (!(c->hash = calloc(c->hash_size, sizeof *c->hash))) {
		free(c->pool);
		free(c);
		return NULL;
	}
	for (i=0; i<c->hash_size; i++) {
		setl(&c->hash[i], NULL, NULL, NULL, NULL);
	}

	for (i=0; i<c->size; i++) {
		p = &c->pool[i];
		setl(p, p-1, p+1, NULL, NULL);
		p->v.size = 0;
		p->v.v = NULL;
	}

	setl(&c->mru, NULL, NULL, NULL, NULL);
	p = c->pool;
	c->mru.link.lru.next = p;
	p->link.lru.prev = &c->mru;

	p = &c->pool[c->size-1];
	p->link.lru.next = NULL;
	c->lru = p;

	return c;
}

static int
wgv_cache_finish(c)
struct wgv_cache *c;
{
	struct cell *p;
	for (p=c->lru; p; p=p->link.lru.next) {
		free(p->v.v);
	}
	free(c->pool);
	free(c->hash);
	free(c);
	return 0;
}

static int
wgv_cache_insert(c, dir, id, v)
struct wgv_cache *c;
idx_t id;
struct xr_vec const *v;
{
	struct cell *p;
	size_t s;

	/* lru kara 1 ko tukamu */
	p = c->lru;
	assert(p!=NULL);
	assert(p!=&c->mru);
/* fprintf(stderr, "wgv_cache_insert: a\n"); */

	/* v wo realloc suru hituyou ga areba, realloc suru */
	if (!p->v.v || p->v.size < v->elem_num) {
/* p->v.v==NULL && v->elem_num==0 no toki ga aru.
   kono baai, newsize = 0 daga, sizeof_dxr_vec(0)>0 nanode,
   realloc ha (0 yori ookii hikisuu de) tadasiku okonawareru.
 */
		struct xr_vec *new;
		ssize_t newsize;
		newsize = NALIGN(v->elem_num, 512);
		if (!(new = realloc(p->v.v, sizeof_dxr_vec(newsize)))) {
			/* realloc ni sippai sitara sonomama return */
			return -1;
		}
		p->v.v = new;
		p->v.size = newsize;
	}
/* fprintf(stderr, "wgv_cache_insert: b, %p, %p, %d, %p\n", p, p->v.v, p->v.size, v); */

	/* p->v ni v wo copy */
	memmove(p->v.v, v, sizeof_dxr_vec(v->elem_num));
	p->key.dir = dir;
	p->key.id = id;
/* fprintf(stderr, "wgv_cache_insert: c\n"); */

	/* lru kara kiri hanasu */
	assert(p->link.lru.next==NULL);
	c->lru = p->link.lru.prev;
	c->lru->link.lru.next = NULL;
/* fprintf(stderr, "wgv_cache_insert: d\n"); */

	/* hash kara kiri hanasu */
	if (p->link.hash.prev) {
		p->link.hash.prev->link.hash.next = p->link.hash.next;
	}
	if (p->link.hash.next) {
		p->link.hash.next->link.hash.prev = p->link.hash.prev;
	}
/* fprintf(stderr, "wgv_cache_insert: e\n"); */

	/* hash ni insert */
	s = hash(c, &p->key);
	p->link.hash.prev = &c->hash[s];
	p->link.hash.next = c->hash[s].link.hash.next;
	if (c->hash[s].link.hash.next) {
		c->hash[s].link.hash.next->link.hash.prev = p;
	}
	c->hash[s].link.hash.next = p;
/* fprintf(stderr, "wgv_cache_insert: f\n"); */

	/* mru ni tuika */
	assert(c->mru.link.lru.prev==NULL);
	p->link.lru.prev = &c->mru;
	p->link.lru.next = c->mru.link.lru.next;
	if (c->mru.link.lru.next) {
		c->mru.link.lru.next->link.lru.prev = p;
	}
	c->mru.link.lru.next = p;
/* fprintf(stderr, "wgv_cache_insert: g\n"); */

	return 0;
}

static ssize_t
wgv_cache_lookup(c, dir, id, v)
struct wgv_cache *c;
idx_t id;
struct xr_vec const **v;
{
	struct cell *p;
	size_t s;
	struct key key;
	key.dir = dir;
	key.id = id;
	s = hash(c, &key);
	for (p=c->hash[s].link.hash.next; p; p=p->link.hash.next) {
		if (p->key.dir == key.dir && p->key.id == key.id) {
			*v = p->v.v;
			return p->v.v->elem_num;
		}
	}
	return -1;
}

static size_t
hash(c, k)
struct wgv_cache *c;
struct key *k;
{
	return (k->dir*19 + k->id*13) % c->hash_size;
}

/** *** **/

ssize_t **
wcm(words, nwords, docs, ndocs, w, dir)
struct syminfo const *words;
struct syminfo const *docs;
WAM *w;
ssize_t nwords, ndocs;
{
	ssize_t **r;
struct timeval start, end;
#if 0
	return wcm0(words, nwords, docs, ndocs, w, dir);
#endif
gettimeofday(&start, NULL);
	if (w->ds.local) {
		if (!XRR(w, dir)) {
			return NULL;
		}
fprintf(stderr, "LOCAL!");
		r = dw_cm(words, nwords, docs, ndocs, w->xr[dir]);
	}
	else {
		struct rcxs *rp;
		ssize_t croot;
fprintf(stderr, "wcm: REMOTE!\n");

		rp = &w->ds.rcx[dir];

		croot = w->param.cascade ? -1 : -2;
fprintf(stderr, "wcm: makesv\n");

		if (!w->s[dir].sv && (w->s[dir].np = makesv(croot, rp->np, rp->rems, &w->s[dir].sv, NULL, w->param.fault_tolerant))==-1) {
			return NULL;
		}
fprintf(stderr, "wcm: rdwcm\n");
#if 0
fprintf(stderr, "wcm: rdwcm\n");
fprintf(stderr, "wcm: rdwcm: w:%p\n", w);
fprintf(stderr, "wcm: rdwcm: s:%p\n", w->s);
fprintf(stderr, "wcm: rdwcm: dir:%d\n", dir);
fprintf(stderr, "wcm: rdwcm: sv:%p\n", w->s[dir].sv);
fprintf(stderr, "wcm: rdwcm: np:%p\n", w->s[dir].np);
#endif

		r = rdwcm(words, nwords, docs, ndocs, w->s[dir].sv, w->s[dir].np, w->param.fault_tolerant);
	}
fprintf(stderr, "wcm: end.\n");
gettimeofday(&end, NULL);
timersub(&end, &start, &end);
fprintf(stderr, "wcm: elapsed time: %ld.%03ld\n", end.tv_sec, end.tv_usec / 1000);
	return r;
}

int
wam_setmask(w, dir, pmask, plen, nmask, nlen)
WAM *w;
idx_t *pmask, *nmask;
size_t plen, nlen;
{
	int e;

	if (w->ds.local) {
		if (!XRR(w, dir)) {
			fprintf(stderr, "!XRR");
			return -1;
		}
		e = dxr_set_mask(w->xr[dir], pmask, plen, nmask, nlen);
		if (e==-1) {
			fprintf(stderr, "!DXRSETMASK\n");
		}
	}
	else {
		struct rcxsp *rp;
		rp = w->ds.rcx[dir].rems[0];	/* use non-cascade rems */
		if ((e = rdsetmask(w, rp, w->ds.rcx[dir].np, pmask, plen, nmask, nlen))==-1) {
			fprintf(stderr, "!RDSETMASK");
			return -1;
		}
	}
	return 0;
}

static int
rdsetmask(w, r, np, pmask, plen, nmask, nlen)
WAM *w;
struct rcxsp *r;
idx_t *pmask, *nmask;
size_t np, plen, nlen;
{
	size_t p;

	for (p=0; p<np; p++) {
		struct rcxsp *rp = &r[p];
		int e;

		if (!rp->f && rcxsp_connect_n_open(rp, w->param.fault_tolerant)==-1) {
			if (w->param.fault_tolerant) {
syslog(LOG_DEBUG, "%s:%s is down. skipped.", rp->host, rp->serv);
				continue;
			}
			return -1;
		}

		if ((e = dwc_dxr_set_mask(rp->f, pmask, plen, nmask, nlen))==-1) {
			fprintf(stderr, __FILN__ ": dwc_dxr_set_mask failed");
			return -1;
		}
	}
	return 0;
}

#if defined GAM_TEST

#include "gam.h"
static int connect_to_gam(void);

static void
report_query(q, qlen, w, dir, weight_type)
struct syminfo const *q;
size_t qlen;
WAM *w;
{
	size_t i;
	int s;
	size_t msg_size;
	struct gam_msg *msg;

	if ((s = connect_to_gam())==-1) {
		return;
	}

	msg_size = sizeof *msg + (qlen-1) * sizeof msg->q[0];

	if (!(msg = calloc(msg_size, 1))) {
		close(s);
		return;
	}
	msg->wam_size = wam_size(w, dir);
	msg->max_elem_num = wam_max_elem_num(w, dir);
	msg->nq = qlen;
	msg->weight_type = weight_type;
	msg->dir = dir;
	strlcpy(msg->handle, wam_handle(w), sizeof msg->handle);
	for (i=0; i<qlen; i++) {
		msg->q[i].id = q[i].id;
		msg->q[i].frq = wam_elem_num(w, dir, q[i].id);
	}
	send(s, msg, msg_size, 0);
	free(msg);
	close(s);
}

static int
connect_to_gam()
{
	char const *host = "localhost";
	char const *serv = GAM_PORT;
	return nnet_cli(host, serv);
}
#endif
