/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwd.c,v 1.6 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>

#include "dwsys.h"
#include "pkt.h"
#include "dxrP.h"
#include "tutil.h"
#include "detache.h"
#include "nnet.h"

#if ! defined SA_RESETHAND
#define SA_RESETHAND	0
#endif

static void unlink_pidf(void);
static void sigflg(int, int, int);
static void sigint(int);
static void sigterm(int);
static void sigchld(int);

static int ss[16];
static int ns;

static char pidf_path[MAXPATHLEN] = "";

char *progname = NULL;

static void
unlink_pidf()
{
	if (*pidf_path) {
		unlink(pidf_path);
	}
}

int
main(argc, argv)
char *argv[];
{
int log2f = 1;
	char *serv = DWD_SERV;
	struct timeval ti;
	time_t t;
	char *date;
	char const *dirname;
	int d;
	char pids[12];
	const char *cause;

	progname = basename(argv[0]);

	openlog(progname, LOG_PID, LOG_LOCAL0);

	if (!strcmp(argv[1], "-E") || !strcmp(argv[1], "-log-to-stderr")) {
		argc--, argv++;
		log2f = 0;
	}

	if (argc != 3) {
		fprintf(stderr, "usage: %s [-E|-log-to-stderr] dirname port\n", progname);
		return 1;
	}

	detachetty();

#if HAVE_IEEEFP_H && defined fpsetprec
	fpsetprec(FP_PE);
#endif

	dirname = argv[1];
	if (argc==3) {
		serv = argv[2];
	}

	nopenlog(progname, LOG_PID, LOG_LOCAL0);
#if ! defined XDEBUG
if (log2f) {
	char log[MAXPATHLEN];
	snprintf(log, sizeof log, "%s" DWD_LOGFILE, dirname);
	nassignlogf(log);
}
#endif
	gettimeofday(&ti, NULL);
	t = ti.tv_sec;
	date = ctime(&t);
	syslog(LOG_DEBUG, "started at %s", date);

	if (chdir(dirname)==-1) {
		syslog(LOG_DEBUG, "%s: %s", dirname, strerror(errno));
		return 1;
	}

	if ((ns = sockets(serv, ss, nelems(ss), &cause)) == 0) {
		syslog(LOG_DEBUG, "sockets: %s", cause);
		return -1;
	}

	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, sigint);
	signal(SIGTERM, sigterm);
	signal(SIGCHLD, sigchld);
	sigflg(SIGCHLD, SA_RESTART, SA_RESETHAND);

	snprintf(pidf_path, sizeof pidf_path, "%s" DWD_PIDF, dirname);

	if ((d = open(pidf_path, O_WRONLY | O_CREAT | O_EXCL, 0666))==-1) {
		syslog(LOG_DEBUG, "%s: %s", pidf_path, strerror(errno));
		return 1;
	}
	sprintf(pids, "%d", getpid());
	if (write(d, pids, strlen(pids))==-1) {
		return 1;
	}
	close(d);
	atexit(unlink_pidf);

	for (;;) {
		int i;
		fd_set readfds;
		int nfds;

		FD_ZERO(&readfds);

		nfds = 0;
		for (i = 0; i < ns; i++) {
			FD_SET(ss[i], &readfds);
			if (ss[i] > nfds) {
				nfds = ss[i];
			}
		}

		nfds++;

		if (select(nfds, &readfds, NULL, NULL, NULL) == -1) {
			if (errno != EINTR) {
				syslog(LOG_DEBUG, "select: %s", strerror(errno));
			}
			continue;
		}

		for (i = 0; i < ns; i++) {
			struct sockaddr_storage addr;
#if ! HAVE_SOCKLEN_T
#define	socklen_t size_t
#endif
			socklen_t len;
			int s;
#if defined TCP_NODELAY
			int one = 1;
#endif
			if (!FD_ISSET(ss[i], &readfds)) {
				continue;
			}
			len = sizeof addr;
			if ((s = accept(ss[i], (struct sockaddr *)&addr, &len)) < 0) {
				syslog(LOG_DEBUG, "accept: %s", strerror(errno));
				continue;
			}

#if 0
/* XXX inet_ntoa			syslog(LOG_DEBUG, "%s: dwd: accept: %s\n", udate(), inet_ntoa(addr.sin_addr)); */
			nlogflush();
#endif

			switch (fork()) {
			case -1:
				syslog(LOG_DEBUG, "fork: %s", strerror(errno));
				break;
			case 0:
#if defined TCP_NODELAY
				(void)setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&one, sizeof one);
#endif
				for (i = 0; i < ns; i++) {
					close(ss[i]);
				}
				dwd_serve(s, dirname);
				shutdown(s, SHUT_RDWR);
				close(s);
				_exit(0);
				break;
			default:
				close(s);
				break;
			}
		}
	}

	return 0;
}

static void
sigflg(sig, set, reset)
{
	struct sigaction sa;
	sigaction(sig, NULL, &sa);
	sa.sa_flags |= set;
	sa.sa_flags &= ~reset;
	sigaction(sig, &sa, NULL);
}

static void
sigint(n)
{
	char *date;
	int i;
	struct timeval ti;
	time_t t;
/*
	if (ss!=-1) {
		shutdown(ss, SHUT_RDWR);
	}
*/
	for (i = 0; i < ns; i++) {
		shutdown(ss[i], SHUT_RDWR);
		if (ss[i] != -1) close(ss[i]);
	}
	gettimeofday(&ti, NULL);
	t = ti.tv_sec;
	date = ctime(&t);
	syslog(LOG_DEBUG, "terminated at %s", date);
	exit(0);
}

static void
sigterm(n)
{
	sigint(n);
}

static void
sigchld(n)
{
	int status;

	for (; wait3(&status, WNOHANG, NULL) > 0; ) ;
}
