/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dxr.c,v 1.18 2007/03/26 23:07:29 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/resource.h>

#include <assert.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/dxr.h>

#include "dfi.h"
#include "dxrP.h"
#include "zvec.h"
#include "gstat.h"

#define NEED_UNCOMPRESS
#include "xcompress.c"
#undef NEED_UNCOMPRESS

struct mask_t {
	idx_t *mask;
	void *buf;
	size_t len, buflen;
	char *bitmask;
};

struct dw_FILE {
	int magic;

	void *p;
	size_t len;

	struct dxr_hdr h, *hp;
	struct dxr_entry *b;
	char const *v;
	struct xr_vec *vt;

/*
	char path[MAXPATHLEN];
*/
	char const *handle;

	struct {
		struct dfilt *s, *d;
	} filt;

	struct mask_t pmask, nmask;
};

static void reset_mask(struct mask_t *);
static int check_dxr_hdr(struct dxr_hdr *);
static void fcopy(struct xr_vec *, struct xr_vec const *, struct dfilt *);
static void mask_vec(struct xr_vec *, idx_t *, size_t, idx_t *, size_t);
static ssize_t dxr_get_vec_sub(DWFILE *, idx_t, struct xr_vec const **, int);
static int set_mask_0(struct mask_t *, idx_t *, size_t);
static int uicompar(const void *, const void *);

DWFILE *
dxr_open(path, handle)
char const *path, *handle;
{
	int d = -1;
	DWFILE *f = NULL;
	struct stat sb;
	struct rlimit rl;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		return NULL;
	}

	f->p = MAP_FAILED;
	f->len = 0;
/*
	strlcpy(f->path, path, sizeof f->path);
*/
	f->handle = handle;
	f->vt = NULL;

	if ((d = open(path, O_RDONLY))==-1) {
		dwerrno = errno;
		goto err;
	}

	if (fstat(d, &sb)==-1) {
		dwerrno = errno;
		goto err;
	}

	f->len = sb.st_size;

#if defined RLIMIT_VMEM
	if (getrlimit(RLIMIT_VMEM, &rl) == -1) {
		perror("getrlimit(RLIMIT_VMEM)");
	}
	else if (rl.rlim_cur < rl.rlim_max) {
		rlim_t prev = rl.rlim_cur;
		rl.rlim_cur = rl.rlim_max;
		if (setrlimit(RLIMIT_VMEM, &rl) == -1) {
			perror("setrlimit(RLIMIT_VMEM)");
		}
		else {
			fprintf(stderr, "RLIMIT_VMEM: %d => %d\n", (int)prev, (int)rl.rlim_cur);
		}
	}
#endif

/*** AAA ***/
	if ((f->p = mmap(0, f->len, PROT_READ, MAP_SHARED, d, 0))==MAP_FAILED) {
		fprintf(stderr, "dxr_open:dxr.c:mmap(,%s,...) => %s\n", sizetostr(f->len), strerror(errno));
		dwerrno = errno;
		goto err;
	}

	close(d);
	d = -1;

	f->hp = f->p;
	memmove(&f->h, f->hp, sizeof f->h);
	f->b = (struct dxr_entry *)((char *)f->p+f->h.entry_table_offset);
	f->v = ((char const *)f->p+f->h.vector_table_offset);
	if (/*f->h.compress_mode!=COMPRESS_NONE &&*/
	    !(f->vt = malloc(sizeof_dxr_vec(f->h.max_elem_num)))) {
		goto err;
	}

	if (check_dxr_hdr(&f->h)==-1) {
		goto err;
	}

	f->filt.s = f->filt.d = NULL;

	reset_mask(&f->pmask);
	reset_mask(&f->nmask);

	f->magic = DXR_MAGIC;
	return f;
err:
	if (d!=-1) {
		close(d);
	}
	dxr_close(f);
	return NULL;
}

static void
reset_mask(m)
struct mask_t *m;
{
	m->mask = m->buf = NULL;
	m->len = m->buflen = 0;
}

static int
check_dxr_hdr(h)
struct dxr_hdr *h;
{
	switch (h->type) {
	case DXR_ROW:
		if (strncmp(h->ident, "!dxr.row", 8)) {
			dwerrno = DWE_BADF;
			return -1;
		}
		break;
	case DXR_COL:
		if (strncmp(h->ident, "!dxr.col", 8)) {
			dwerrno = DWE_BADF;
			return -1;
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (h->version != DXR_VERSION) {
		dwerrno = DWE_BADVER;
		return -1;
	}
	return 0;
}

int
dxr_close(f)
DWFILE *f;
{
	int e = 0;
	f->magic = 0;
	if (f->p!=MAP_FAILED) {
		e = munmap(f->p, f->len);
	}
	free(f->pmask.buf);
	free(f->nmask.buf);
	free(f->vt);
	free(f);
	return e;
}

int
dxr_mtime(f)
DWFILE *f;
{
	return f->h.mtime;
}

ssize_t
dxr_size(f)
DWFILE *f;
{
	return f->h.nentries;
}

int
dxr_need_dup_p(f)
DWFILE *f;
{
	return f->h.compress_mode!=COMPRESS_NONE || f->filt.d || f->pmask.mask || f->nmask.mask;
}

#define	IDCHK(f, id, r) do { \
	struct dfilt *fi; \
	if ((id)==0 || (id)>(f)->h.nentries) { \
		dwerrno = DWE_BADR; \
		return (r); \
	} \
	if ((fi = (f)->filt.s) && (*fi->screened)(fi->cookie, (id))) { \
		dwerrno = DWE_BADR; \
		return (r); \
	} \
} while (0)

static void
fcopy(u, v, f)
struct xr_vec *u;
struct xr_vec const *v;
struct dfilt *f;
{
	size_t i;
	struct xr_elem *q;
	ssize_t elem_num;
	freq_t freq_sum;

	elem_num = v->elem_num;
	q = u->elems;
	freq_sum = 0;
	for (i=0; i<elem_num; i++) {
		idx_t id = v->elems[i].id;
		if (!(*f->screened)(f->cookie, id)) {
			q->id = id;
			q->freq = v->elems[i].freq;
			freq_sum += q->freq;
			q++;
		}
	}
	u->elem_num = q-u->elems;
	u->freq_sum = freq_sum;
}

static void
mask_vec(v, pmask, plen, nmask, nlen)
struct xr_vec *v;
idx_t *pmask, *nmask;
size_t plen, nlen;
{
	idx_t *pme = NULL, *nme = NULL;
	struct xr_elem *p, *q, *pe;
	freq_t freq_sum;

	if (pmask) {
		/* assert(plen); */
		pme = pmask + plen;
	}
	if (nmask) {
		/* assert(nlen); */
		nme = nmask + nlen;
	}
	freq_sum = 0;
	for (p = q = v->elems, pe = p + v->elem_num; p < pe; p++) {
		idx_t id = p->id;
		if (pmask) for (; pmask < pme && (*pmask) < id; pmask++) ;
		if (nmask) for (; nmask < nme && (*nmask) < id; nmask++) ;
		if ((pmask && pmask < pme && (*pmask) == id) ||
		    (nmask && (nmask >= nme || (*nmask) != id))) {
			q->id = id;
			q->freq = p->freq;
			freq_sum += q->freq;
			q++;
		}
	}
	v->elem_num = q - v->elems;
	v->freq_sum = freq_sum;
}

ssize_t
dxr_get_vec(f, id, v)
DWFILE *f;
idx_t id;
struct xr_vec const **v;
{
	ssize_t e;
/* fprintf(stderr, "dxr_get_vec: %p %u %p\n", f, id, v); */
	e = dxr_get_vec_sub(f, id, v, 0);
/* fprintf(stderr, "dxr_get_vec: return %d\n", e); */
	return e;
}

ssize_t
dxr_get_vec_dup(f, id, v)
DWFILE *f;
idx_t id;
struct xr_vec const **v;
{
	ssize_t e;
/* fprintf(stderr, "dxr_get_vec_dup: %p %u %p\n", f, id, v); */
	/* duplicate==1 nara, v ha kakikomi kanou. warning ha musi suru koto. */
	e = dxr_get_vec_sub(f, id, v, 1);
/* fprintf(stderr, "dxr_get_vec_dup: return %d\n", e); */
	return e;
}

static ssize_t
dxr_get_vec_sub(f, id, v, duplicate)
DWFILE *f;
idx_t id;
struct xr_vec const **v;
{
	struct dxr_entry *fp;
	void const *addr;
	size_t l;
	struct xr_vec *vt = NULL;
/* fprintf(stderr, "dxr_get_vec_sub: %p %u %p %d\n", f, id, v, duplicate); */

	IDCHK(f, id, -1);
	fp = &f->b[id];
	addr = f->v+fp->addr;

#ifdef GSTAT
{
	size_t l = fp->addr%(8*1024);
	if (l<nelems(gsb->xrpage)) {
		gsb->xrpage[l]++;
	}
}
#endif

	l = sizeof_dxr_vec(fp->elem_num);
	if (duplicate) {
		if (!(vt = malloc(l))) {
			dwerrno = errno;
/* fprintf(stderr, "dxr_get_vec_sub: failed: !malloc\n"); */
			return -1;
		}
	}
	else {
		vt = f->vt;
	}
/* fprintf(stderr, "dxr_get_vec_sub: A: %p\n", vt); */
	if (f->h.compress_mode==COMPRESS_NONE) {
/* fprintf(stderr, "dxr_get_vec_sub: B1\n"); */
		if (f->filt.d) {
			fcopy(vt, addr, f->filt.d);
		}
		else if (duplicate || f->pmask.mask || f->nmask.mask) {
			memmove(vt, addr, l);
		}
		else {
	/* duplicate==1 nara, koko ha toora nai. warning ha musi suru koto. */
			vt = addr;
		}
	}
	else {
/* fprintf(stderr, "dxr_get_vec_sub: B2\n"); */
		vt->elem_num = fp->elem_num;
		vt->freq_sum = fp->freq_sum;
		if (dxr_uncompress_vec(vt, addr, f->h.compress_mode, f->filt.d)==-1) {
			dwerrno = DWE_IERROR;
			if (duplicate) {
				free(vt);
			}
/* fprintf(stderr, "dxr_get_vec_sub: failed: !uncompress vec\n"); */
			return -1;
		}
	}
/* fprintf(stderr, "dxr_get_vec_sub: C\n"); */
	if (f->pmask.mask || f->nmask.mask) {
		mask_vec(vt, f->pmask.mask, f->pmask.len, f->nmask.mask, f->nmask.len);
	}
	*v = vt;
/* fprintf(stderr, "dxr_get_vec_sub: return %d\n", vt->elem_num); */
	return vt->elem_num;
}

ssize_t
dxr_elem_num(f, id)
DWFILE *f;
idx_t id;
{
	IDCHK(f, id, -1);
	return f->b[id].elem_num;
}

freq_t
dxr_freq_sum(f, id)
DWFILE *f;
idx_t id;
{
	IDCHK(f, id, -1);
	return f->b[id].freq_sum;
}

ssize_t
dxr_partition(f, id)
DWFILE *f;
idx_t id;
{
	IDCHK(f, id, -1);
	return f->b[id].addr;
}

ssize_t
dxr_max_elem_num(f)
DWFILE *f;
{
	return f->h.max_elem_num;
}

freq_t
dxr_max_freq_sum(f)
DWFILE *f;
{
	return f->h.max_freq_sum;
}

freq_t
dxr_total_elem_num(f)
DWFILE *f;
{
	return f->h.total_elem_num;
}

freq_t
dxr_total_freq_sum(f)
DWFILE *f;
{
	return f->h.total_freq_sum;
}

struct xr_vec *
dxr_dup(p)
struct xr_vec const *p;
{
	struct xr_vec *q;
	size_t l;

	l = sizeof_dxr_vec(p->elem_num);
	if (!(q = malloc(l))) {
		dwerrno = errno;
		return NULL;
	}
	memmove(q, p, l);
	return q;
}

int
dxr_setf(f, s, d)
DWFILE *f;
struct dfilt *s, *d;
{
	f->filt.s = s;
	f->filt.d = d;

	if (d && !f->vt && !(f->vt = malloc(sizeof_dxr_vec(f->h.max_elem_num)))) {
		return -1;
	}
	return 0;
}

char const *
dxr_handle(f)
DWFILE *f;
{
	return f->handle;
}

char const *
dxr_compress_type(f)
DWFILE *f;
{
	return dw_decode_compress(f->h.compress_mode);
}

int
dxr_set_mask(f, pmask, plen, nmask, nlen)
DWFILE *f;
idx_t *pmask, *nmask;
size_t plen, nlen;
{
	if (set_mask_0(&f->pmask, pmask, plen) == -1) {
		return -1;
	}
	if (set_mask_0(&f->nmask, nmask, nlen) == -1) {
		return -1;
	}
	return 0;
}

static int
set_mask_0(m, mask, len)
struct mask_t *m;
idx_t *mask;
size_t len;
{
	if (!mask) {
		m->mask = NULL;
	}
	else {
		size_t i, reqsize;
		reqsize = len * sizeof *mask;
		if (m->buflen < reqsize + 1) {
			void *new;
			if (!(new = realloc(m->buf, reqsize + 1))) {
				return -1;
			}
			m->buflen = reqsize + 1;
			m->buf = new;
		}
		m->mask = m->buf;
		if (len) memmove(m->mask, mask, reqsize);
		m->len = len;
		if (len) for (i=0; i<len-1; i++) {
			if (!(m->mask[i] <= m->mask[i+1])) {
				qsort(m->mask, len, sizeof *m->mask, uicompar);
				break;
			}
		}
/*
fprintf(stderr, "set_mask_0: ");
prntilst(stderr, m->mask, len);
fprintf(stderr, "\n");
*/
	}
	return 0;
}

static int
uicompar(va, vb)
const void *va, *vb;
{
	const idx_t *a, *b;
	a = va;
	b = vb;
	if ((*a)<(*b)) {
		return -1;
	}
	else if ((*a)>(*b)) {
		return 1;
	}
	return 0;
}

/*
prntilst(stream, l, len)
FILE *stream;
size_t *l;
{
	size_t i;
	fprintf(stream, "[");
	for (i=0; i<len; i++) {
		fprintf(stream, "%d", l[i]);
		if (i<len-1) fprintf(stream, " ");
	}
	fprintf(stream, "], %d", len);
}
*/
