/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: mkpi.c,v 1.6 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <assert.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>

#include "dwsys.h"
#include "dxrP.h"
#include "dauiP.h"

char *progname;

struct source {
	RFILES *f;
	DWFILE *e;
	DWFILE *c;
};

#if defined X_READ_RAW_VEC

struct xread_param {
	struct {
		int m;
		double p;
	} a, w;
};

struct xread_param *xrp = NULL;

#endif

static void usage(void);
static int toplevel(char const *, char const *, char const *, char const *, int);
static int mkaui(char const *, int, struct source *f, char const *);
static int proc_rvec(struct source *, ssize_t, struct dxr_raw_vec *, struct au_index_t *, ssize_t);
static int daui_write(struct au_index_t *, ssize_t, ssize_t, int, FILE *);
static void setup_o(char const *, char const *, char const *, int);
static void start_o(void);
static ssize_t finish_o(void);
static int open_o(void);
static int close_o(void);
static int put_entry(struct dxr_raw_vec *, ssize_t *, size_t *, size_t *);
static int splitp(void);
static ssize_t src_read_raw_vec(struct source *, struct dxr_raw_vec **, idx_t);

#if defined X_READ_RAW_VEC
static int parse_xread_param(struct xread_param *, char *);

static int
parse_xread_param(xp, s)
struct xread_param *xp;
char *s;
{
	if (!*s) return -1;
	xp->a.m = strtol(s, &s, 10);
	if (*s!=',') return -2;
	if (!*++s) return -3;
	xp->a.p = strtod(s, &s);
	if (*s!=',') return -4;
	if (!*++s) return -5;
	xp->w.m = strtol(s, &s, 10);
	if (*s!=',') return -6;
	if (!*++s) return -7;
	xp->w.p = strtod(s, &s);
	if (*s) return -8;
	return 0;
}
#endif

static void
usage()
{
	fprintf(stderr, "usage: %s [-h] "
#if defined X_READ_RAW_VEC
		"[-xread-param #,#,#,#] "
#endif
		"dirname handle [row|col] aux_basename file|@files|!command\n", progname);
	exit(1);
}

int
main(argc, argv)
char *argv[];
{
	int mtime;
	struct timeval start_t;
	char *direction;
	int dir;

	char *dirname, *handle, *abase, *files;
	char dcw_path[MAXPATHLEN];
	char aui_path[MAXPATHLEN];

	progname = basename(argv[0]);
	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(*argv, "-h")) {
			usage();
		}
#if defined X_READ_RAW_VEC
		else if (!strcmp(*argv, "-xread-param") && argc>1) {
			static struct xread_param xrp0;
			int e;
			argc--;
			argv++;
			if ((e=parse_xread_param(&xrp0, *argv))<0) {
				fprintf(stderr, "parse_xread_param failed for [%s], %d\n", *argv, e);
				usage();
			}
			xrp = &xrp0;
		}
#endif
		else {
			break;
		}
	}
	if (argc!=5) {
		usage();
	}

	dirname = argv[0];
	handle = argv[1];
	direction = argv[2];
	abase = argv[3];
	files=argv[4];

	if (!strcasecmp(direction, "row")) {
		dir = DCW_ROW;
	}
	else if (!strcasecmp(direction, "col")) {
		dir = DCW_COL;
	}
	else {
		usage();
	}

	dw_mkpathname(dcw_path, sizeof dcw_path, dirname, handle, dir, 0, 1);
	dw_mkauipath(aui_path, sizeof aui_path, dirname, handle, abase, -1, 0);

#if defined ANON_MTIME
	mtime = ~0;
#else
	gettimeofday(&start_t, NULL);
	mtime = start_t.tv_sec;
#endif

	setup_o(dirname, handle, abase, mtime);

	if (toplevel(handle, files, aui_path, dcw_path, mtime)==-1) {
		fprintf(stderr, "%s failed\n", progname);
		return 1;
	}
	return 0;
}

static int
toplevel(handle, files, aui_path, dcw_path, mtime)
char const *handle, *files, *aui_path, *dcw_path;
{
	struct source so;
	struct timeval start, curr, elapsed;

	if (!(so.c = dcw_open(dcw_path, handle))) {
		nperror(dcw_path);
		return -1;
	}

	so.e = NULL;
	so.f = NULL;

	if (*files=='!') {
		char const *cmd = files+1;
		if (!(so.e=dau_open_cmd(handle, cmd, so.c))) {
			nperror("dau_open_cmd");
			return -1;
		}
	}
	else {
		char **rfiles = NULL;
		if (!(rfiles = read_files(files))) {
			fprintf(stderr, "%s: read_files failed\n", progname);
			return 1;
		}
		if (!(so.f=init_read_raw_vec(rfiles))) {
			nperror("init_read_raw_vec");
			return -1;
		}

#if defined X_READ_RAW_VEC
		if (xrp) {
			fprintf(stderr, "xrp: %d %f %d %f\n", xrp->a.m, xrp->a.p, xrp->w.m, xrp->w.p);
			rawvec_set_xread_param(so.f, xrp->a.m, xrp->a.p, xrp->w.m, xrp->w.p);
		}
#endif
	}

	gettimeofday(&curr, NULL);
	start = curr;

	if (mkaui(handle, mtime, &so, aui_path)==-1) {
		fprintf(stderr, "%s: mkdcw failed\n", progname);
		return -1;
	}

	gettimeofday(&curr, NULL);
	timersub(&curr, &start, &elapsed);
	printf("elapsed time: " TIMERFMT "\n", TIMERSCL(&elapsed));

	if (so.f) {
		finish_read_raw_vec(so.f);
	}

	return 0;
}

static int
mkaui(handle, mtime, so, aui_path)
char const *handle;
struct source *so;
char const *aui_path;
{
	FILE *g;
	ssize_t e=0;
	struct dxr_raw_vec *rv;
	ssize_t n, filled, pe;

	struct au_index_t *t;

	start_o();

	n = dcw_size(so->c);

	if (!(t = calloc(n+1, sizeof *t))) {
		nperror("calloc t");
		return -1;
	}

	if (so->f) {
		report_progress_raw_vec(so->f, stdout, 0);

		if (rewind_read_raw_vec(so->f)==-1) {
			nperror("rewind_raw_vec");
			return -1;
		}
	}

	if (so->f) {
		for (filled = 0; (e=src_read_raw_vec(so, &rv, 0))>=0; filled++) {
			if (proc_rvec(so, e, rv, t, n)==-1) {
				break;
			}
		}
	}
	else {
		idx_t id;
		for (filled = 0, id = 1; id<=n; id++, filled++) {
			if ((e = src_read_raw_vec(so, &rv, id))<0) {
				break;
			}
			if (proc_rvec(so, e, rv, t, n)==-1) {
				break;
			}
		}
	}
	printf("\n");
	if (filled != n) {
		fprintf(stderr, "%s: warning: title file may be shorter than frq file\n", progname);
	}

	if (e==-1) {
		fprintf(stderr, "%s: file read error\n", progname);
		return -1;
	}
	if ((pe = finish_o())==-1) {
		return -1;
	}
	if (!(g = fopen(aui_path, "w"))) {
		nperror(aui_path);
		return -1;
	}
	if (daui_write(t, n, pe, mtime, g)==-1) {
		fprintf(stderr, "%s: write index file failed\n", progname);
		return -1;
	}
	fclose(g);
	return 0;
}

static int
proc_rvec(so, e, rv, t, size)
struct dxr_raw_vec *rv;
struct source *so;
struct au_index_t *t;
ssize_t e;
ssize_t size;
{
	ssize_t frag;
	size_t addr;
	size_t elem_len;
	if (e==0) {
		fprintf(stderr, "\nwarning: empty document: [%s]\n", rv->key);
	}
	switch (rv->ftype) {
		idx_t id;
		struct au_index_t *p;
		
	case RAWXR_TYPE_ASCII:
	case RAWXR_TYPE_BINARY:
		if ((id = dcw_name2id(so->c, rv->key)) == 0) {
			fprintf(stderr, "%s: warning: unrelated title entry: key=[%s]\n", progname, rv->key);
		}
		if (size<id) {
			fprintf(stderr, "%s: error: title id out of range key=[%s]\n", progname, rv->key);
			return -1;
		}
		p = &t[id];
#if 0
fprintf(stderr, 
"key=[%s], id=%d, frag=%d, off=%ld, keylen=%d, elemlen=%d\n", 
rv->key, id, rv->frag, rv->pos, rv->key_len, rv->elem_len);
#endif
		if (p->offset) {
			fprintf(stderr, "%s: warning: duplicated title entry: key=[%s]\n", progname, rv->key);
		}
		if (put_entry(rv, &frag, &addr, &elem_len)==-1) {
			return -1;
		}
		p->fragment = frag;
		p->offset = addr;
		p->length = elem_len;

		break;
	default:
		fprintf(stderr, "%s: invalid file type: %c\n", progname, rv->ftype);
		return -1;
		break;
	}
	return 0;
}

static int
daui_write(t, n, p, mtime, f)
struct au_index_t *t;
ssize_t n, p;
FILE *f;
{
	struct daui_hdr h;

	if (fill_daui_hdr(&h, mtime, -1)==-1) {
		return -1;
	}
	h.nentries = n;
	h.pentries = p;

	if (fwrite(&h, sizeof h, 1, f)!=1) {
		return -1;
	}
	if (fwrite(t, sizeof *t, n+1, f)!=n+1) {
		return -1;
	}

	return 0;
}

struct {
	FILE *f;
	size_t addr;
	ssize_t frag;
	int mtime;
	ssize_t nentries, pentries;

	char dirname[MAXPATHLEN];
	char handle[MAXPATHLEN];
	char abase[MAXPATHLEN];
} o;

static void
setup_o(dirname, handle, abase, mtime)
char const *dirname, *handle, *abase;
{
	strlcpy(o.dirname, dirname, sizeof o.dirname);
	strlcpy(o.handle, handle, sizeof o.handle);
	strlcpy(o.abase, abase, sizeof o.abase);
	o.mtime = mtime;
}

static void
start_o()
{
	o.f = NULL;
	o.frag = 0;
	o.pentries = 0;
	o.addr = 0;
	o.nentries = 0;
}

static ssize_t
finish_o()
{
	if (o.f && close_o()==-1) {
		return -1;
	}
	return o.pentries;
}

static int
open_o()
{
	char path[MAXPATHLEN];
	struct daui_hdr h;

	dw_mkauipath(path, sizeof path, o.dirname, o.handle, o.abase, o.frag, 0);
	if (!(o.f = fopen(path, "w"))) {
		nperror(path);
		return -1;
	}

	bzero(&h, sizeof h);
	if (fwrite(&h, sizeof h, 1, o.f)!=1) {
		return -1;
	}

	o.addr = sizeof h;
	o.nentries = 0;
	return 0;
}

static int
close_o()
{
	struct daui_hdr h;
	if (o.frag==0) {
		o.pentries = o.nentries;
	}

	if (fill_daui_hdr(&h, o.mtime, o.frag)==-1) {
		return -1;
	}
	h.nentries = o.nentries;
	h.pentries = o.pentries;
	rewind(o.f);
	if (fwrite(&h, sizeof h, 1, o.f)!=1) {
		return -1;
	}

	o.frag++;
	o.f = NULL;
	return 0;
}

/* rv->ftype ga BINARY no toki ni, addrp no (zikai no) zoubun to, sizep ha kotonaru atai ni naru koto ni tyuui. */
static int
put_entry(rv, fragp, addrp, sizep)
struct dxr_raw_vec *rv;
ssize_t *fragp;
size_t *addrp;
size_t *sizep;
{
	size_t size;
	if (o.frag==0 && splitp()) {
		if (o.nentries==0 || !o.f) {
			fprintf(stderr, "configuratin error? try to create empty portion\n");
			return -1;
		}
		if (close_o()==-1) {
			return -1;
		}
	}
	else if (o.frag!=0 && o.nentries == o.pentries) {
		if (close_o()==-1) {
			return -1;
		}
	}
	if (!o.f && open_o()==-1) {
		return -1;
	}
	size = 0;
	*addrp = o.addr;
	*fragp = o.frag;
	switch (rv->ftype) {
		size_t l;
		int pad;
		size_t i;
	case RAWXR_TYPE_ASCII:
		for (i=0; i<rv->elem_num; i++) {
			char const *p;
			p = rv->elems[i].name;
			l = strlen(p);
			if (fwrite(p, 1, l, o.f)!=l) {
				return -1;
			}
			o.addr += l;
			size += l;
		}
		break;
	case RAWXR_TYPE_BINARY:
		/* XXX hence rawvec does not support BINARY, this code is unreacheable */
		for (i=0; i<rv->elem_num; i++) {
			char const *p;
			p = rv->elems[i].name;
			l = rv->elems[i].tf;
			if (fwrite(p, 1, l, o.f)!=l) {
				return -1;
			}
			o.addr += l;
			size += l;
		}

		l = sizeof pad - size%sizeof pad;
		pad = 0;
		if (fwrite(&pad, 1, l, o.f)!=l) {
			return -1;
		}
		o.addr += l;
		/* kokodeha, size ha henka sinai */

		break;
	default:
		fprintf(stderr, "configulation error: unsupported ftype\n");
		return -1;
	}
	*sizep = size;
	o.nentries++;
	return 0;
}

static int
splitp()
{
	return o.addr>=512*1024*1024;
}

static ssize_t
src_read_raw_vec(so, rvp, id)
struct source *so;
struct dxr_raw_vec **rvp;
idx_t id;
{
	if (so->f) {
		return read_raw_vec(so->f, rvp);
	}
	else {
		char const *p, *k;
		static struct dxr_raw_vec rv;
		size_t e;
		if (!(k = dcw_id2name(so->c, id))) {
			return -1;
		}
		if ((e = dau_gets(so->e, id, &p)) ==-1) {
			return -1;
		}
		rv.key = k;
		rv.key_len = strlen(rv.key);
		rv.ftype = RAWXR_TYPE_BINARY;
		rv.pos = 0;
		rv.frag = 0;
		rv.elem_num = 1;
		rv.elem_len = e;
		rv.elems[0].name = p;
		rv.elems[0].tf = e;	/* 1 ni subeki ? */
		*rvp = &rv;
		return e;
	}
}
