/*
 *	Copyright (c) 2004 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: nht.c,v 1.9 2007/03/24 00:32:56 nis Exp $";
#endif

#include <config.h>

#include <sys/param.h>

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>

#include "dwsys.h"
#include "sdup.h"
#include "nht.h"
#include "mmalloc.h"
#include "nsxhash.h"

static int nht_extend(struct ht *);
static int h_rehash(struct ht *);
static void ht_stat(struct ht *, FILE *);

#define	DIRTY	IDX_T_MSB1

/* #define	NHTFRGSZ	(1024 * 1024)		/* 1M entry / fragment */
#define	NHTFRGSZ	(128 * 1024)		/* 128k entry / fragment */

#define	BASES_SIZE	((1024 * 1024 * 1024) / NHTFRGSZ) /* total 1G entry */

struct ht *
nht_init(addr)
void *addr;
{
	struct ht *h;

	if (!(h = calloc(1, sizeof *h))) {
		return NULL;
	}
	h->size = 0;
	h->nentries = 0;
	h->threshold = THRESHOLD(h->size);
	h->base = h->pend = addr;
	h->nbases = 0;
	h->monolithic = addr != NULL;
	switch (h->monolithic) {
	case 0:
		h->bases_size = BASES_SIZE;
		if (!(h->bases = calloc(h->bases_size, sizeof *h->bases))) {
			free(h);
			return NULL;
		}
		break;
	case 1:
		h->bases_size = 0;
		h->bases = NULL;
		break;
	default:
		fprintf(stderr, "nht: internal error\n");
		return NULL;
	}
	h->fragsz = NHTFRGSZ;
	return h;
}

static int
nht_extend(h)
struct ht *h;
{
	struct hte *p;
	size_t i;

	switch (h->monolithic) {
	case 0:
		if (h->nbases >= h->bases_size) {
			fprintf(stderr, "nht_extend: error: too many frags\n");
			return -1;
		}
		if (!(p = malloc(h->fragsz * sizeof *p))) {
			return -1;
		}
		h->bases[h->nbases++] = p;
		h->pend = p + h->fragsz;
		break;
	case 1:
		if (!(p = mmalc(h->pend, h->fragsz * sizeof *p))) {
			return -1;
		}
		if (p != h->pend) {
			fprintf(stderr, "nht_extend failed: (!continuous)\n");
			return -1;
		}
		h->pend += h->fragsz;
		break;
	case 2:
		fprintf(stderr, "nht_extend failed: !extensible\n");
		return -1;
	default:
		fprintf(stderr, "nht: internal error\n");
		return -1;
	}
	for (i = 0; i < h->fragsz; i++) {
		struct hte *e = &p[i];
		e->name.p = NULL;
	}
	h->size += h->fragsz;
	return 0;
}

void
nht_free(h)
struct ht *h;
{
	struct hte *p;
	size_t i;

	switch (h->monolithic) {
	case 0:
		for (i = 0; i < h->nbases; i++) {
			free(h->bases[i]);
		}
		free(h->bases);
		break;
	case 1:
		for (p = h->base; p < h->pend; p += h->fragsz) {
			mmfree(p, h->fragsz * sizeof *p);
		}
		break;
	case 2:
		free(h->base);
		break;
	default:
		fprintf(stderr, "nht: internal error\n");
		break;
	}
	free(h);
}

size_t
nht_nent(h)
struct ht *h;
{
	return h->nentries;
}

struct hte *
nht_enter(h, p, sd)
struct ht *h;
char const *p;
struct sdup *sd;
{
	hash_t s;
	size_t c;

	if (h->threshold <= h->nentries) {
		if (h_rehash(h) == -1) {
			return NULL;
		}
	}

	c = 0;
	for (s = nsxhash(h->size, p, c++); ; s = nsxhash(h->size, p, c++)) {
		struct hte *e = hta(h, s);
		if (!e->name.p) {	/* new */
			h->nentries++;
			if (!(e->name.p = sdup(sd, p))) {
				return NULL;
			}
			e->i.tf = 0;
			e->i.df = 0;
			e->i.id = 0;
			return e;
		}
		else if (!strcmp(e->name.p, p)) {	/* exists */
			return e;
		}
		/* else conflict */
	}
	/* notreached */
}

static int
h_rehash(h)
struct ht *h;
{
	hash_t s;
	size_t c;
	size_t osize;
	size_t i, m;
	size_t cnf;

	osize = h->size;
	if (nht_extend(h) == -1) {
		return -1;
	}

	for (i = 0, m = 0; i < osize; i++) {
		struct hte *o = hta(h, i);
		if (o->name.p) {
			o->i.id |= DIRTY;
			m++;
		}
	}
	assert(m == h->nentries);
	h->stat.max = h->stat.total = 0;
	for (i = 0; i < nelems(h->stat.dist); i++) {
		h->stat.dist[i] = 0;
	}
	for (i = 0, m = 0; i < osize; i++) {
		struct hte *o = hta(h, i);
		if (o->name.p && (o->i.id & DIRTY)) {
			char *p = o->name.p;
			o->i.id &= ~DIRTY;
			m++;
			cnf = 0;
			c = 0;
			for (s = nsxhash(h->size, p, c++); ; s = nsxhash(h->size, p, c++)) {
				struct hte *e = hta(h, s);
				if (!e->name.p) {	/* empty */
					*e = *o;
					o->name.p = NULL;
					o->i.tf = o->i.df = 0;	/* redundant */
					break;
				}
				else if (!strcmp(e->name.p, p)) {	/* self */
				/* 	assert(o == e); */
					assert(i == s);
					break;
				}
				else if (e->i.id & DIRTY) {	/* dirty */
					struct hte t;
				/*	assert(o < e); */
					assert(i < s);
					t = *e;
					*e = *o;
					*o = t;
					i--;
					break;
				}
				/* else conflict */
				cnf++;
			}
			if (h->stat.max < cnf) {
				h->stat.max = cnf;
			}
			h->stat.total += cnf;
			if (cnf < nelems(h->stat.dist) - 1) {
				h->stat.dist[cnf]++;
			}
			else {
				h->stat.dist[nelems(h->stat.dist) - 1]++;
			}
		}
	}
	assert(m == h->nentries);
	if (h->nentries) {
		ht_stat(h, stderr);
	}
	h->threshold = THRESHOLD(h->size);
	return 0;
}

static void
ht_stat(h, stream)
struct ht *h;
FILE *stream;
{
	size_t i;
	fprintf(stream, "\nrehash: \n");
	fprintf(stream, "size: %s\n", sizetostr(h->size / sizeof *hta(h, 0)));
	fprintf(stream, "nent: %s\n", sizetostr(h->nentries));
	fprintf(stream, "dens: %f%%\n", (double)h->nentries / h->size * 100);
	fprintf(stream, "ncfl: %s %f\n", sizetostr(h->stat.max), (double)h->stat.total / h->nentries);
	for (i = 0; i < nelems(h->stat.dist); i++) {
		fprintf(stream, "%s: %s (%f%%)\n", sizetostr(i), sizetostr(h->stat.dist[i]), (double)h->stat.dist[i] / h->nentries * 100);
	}
}

struct hte *
nlt_append(h, p, sd)
struct ht *h;
char const *p;
struct sdup *sd;
{
	struct hte *e;

	if (h->size <= h->nentries) {
		if (nht_extend(h) == -1) {
			return NULL;
		}
	}

	e = hta(h, h->nentries);
	h->nentries++;
	if (e->name.p) {
		fprintf(stderr, "lt_append: internal error\n");
		return NULL;
	}

	e->i.tf = 0;
	e->i.df = 0;
	e->i.id = 0;

	if (!(e->name.p = sdup(sd, p))) {
		return NULL;
	}
	return e;
}

void
nht_compact(h, new_nentries, kill_zero)
struct ht *h;
size_t *new_nentries;
{
	size_t i, j, k;

	for (i = 0, j = 0, k = 0; i < h->size; i++) {
		struct hte *o = hta(h, i);
		if (o->name.p) {
			if (!kill_zero || o->i.tf) {
				if (i != j) {
					struct hte *e = hta(h, j);
					*e = *o;
				}
				j++;
			}
			else {
				k++;
			}
		}
	}
	assert(j + k == h->nentries);
#if defined XXDEBUG
	fprintf(stderr, "%s: j: %s k: %s\n", udate(), sizetostr(j), sizetostr(k));
#endif
	*new_nentries = j;
	for (; j < h->size; j++) {
		struct hte *e = hta(h, j);
		e->name.p = NULL;
		e->i.tf = e->i.df = 0;
	}
}

struct hte *
_hta(h, i)
struct ht *h;
size_t i;
{
	assert(h->monolithic == 0);
	return &h->bases[i / h->fragsz][i % h->fragsz];
}

int
nht_sort(h, nmemb, compar)
struct ht *h;
size_t nmemb;
int (*compar)(const void *, const void *);
{
	struct hte x;
	size_t i, j;
	size_t l, r;

	if (nmemb == 0) {
		return 0;
	}

/* XXX l = nmemb / 2 de yoi? */
	l = (nmemb + 1) / 2;
	r = nmemb - 1;

	while (0 < l) {
		l--;
		for (i = l, x = *hta(h, i); (j = 2 * i + 1) <= r; i = j) {
			if (j + 1 <= r && (*compar)(hta(h, j), hta(h, j + 1)) < 0) {
				j++;
			}
			if ((*compar)(hta(h, j), &x) < 0) {
				break;
			}
			*hta(h, i) = *hta(h, j);
		}
		if (i != l) {
			*hta(h, i) = x;
		}
	}

	while (0 < r) {
		x = *hta(h, r);
		*hta(h, r) = *hta(h, 0);
		r--;

		for (i = 0; (j = 2 * i + 1) <= r; i = j) {
			if (j + 1 <= r && (*compar)(hta(h, j), hta(h, j + 1)) < 0) {
				j++;
			}
			if ((*compar)(hta(h, j), &x) < 0) {
				break;
			}
			*hta(h, i) = *hta(h, j);
		}
		*hta(h, i) = x;
	}
	return 0;
}

void
nht_refresh(h)
struct ht *h;
{
	struct hte *p;
	size_t i;

	switch (h->monolithic) {
	case 0:
		for (i = 0; i < h->nbases; i++) {
			p = h->bases[i];
			refresh_mem(p, h->fragsz * sizeof *p);
		}
		break;
	case 1:
		for (p = h->base; p < h->pend; p += h->fragsz) {
			refresh_mem(p, h->fragsz * sizeof *p);
		}
		break;
	case 2:
		refresh_mem(h->base, h->size * sizeof *h->base);
		break;
	default:
		fprintf(stderr, "nht: internal error\n");
		break;
	}
}
