/* this code is derived from NetBSD manpage getaddrinfo(3) */

#include <config.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "nnet.h"

#if ! defined MIN
#define	MIN(a, b)	((a) < (b) ? (a) : (b))
#endif
#define nelems(e)	(sizeof (e) / sizeof *(e))

int nnet_errno;

int
nnet_cli(host, serv)
char const *host;
char const *serv;
{
	struct addrinfo hints, *res, *res0;
	int error;
	int s;
	const char *cause = NULL;

	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = 0;
	hints.ai_protocol = 0;
	hints.ai_addrlen = 0;
	hints.ai_canonname = NULL;
	hints.ai_addr = NULL;
	hints.ai_next = NULL;
	nnet_errno = 0;
	error = getaddrinfo(host, serv, &hints, &res0);
	if (error) {
		syslog(LOG_ERR, "%s\n", gai_strerror(error));
		return -1;
	}
	s = -1;
	for (res = res0; res; res = res->ai_next) {
		s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s < 0) {
			cause = "socket";
			nnet_errno = errno;
			continue;
		}

		if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
			cause = "connect";
			nnet_errno = errno;
			close(s);
			s = -1;
			continue;
		}

		break;  /* okay we got one */
	}
	if (s < 0) {
		syslog(LOG_ERR, "%s\n", cause);
	}
	else {
		nnet_errno = 0;
	}
	freeaddrinfo(res0);
	return s;
}

#define	BACKLOG	128

int
nnet_srv(serv, data, servname)
u_int (*serv)(int, void *, struct sockaddr_storage *);
void *data;
char const *servname;
{
	struct timeval timeout, *tos;
	u_int to;
	struct addrinfo hints, *res, *res0;
	int error;
	int s[8];
	int nsock;
	const char *cause = NULL;
	fd_set readfds;
	int nfds;
	int ss;

	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_protocol = 0;
	hints.ai_addrlen = 0;
	hints.ai_canonname = NULL;
	hints.ai_addr = NULL;
	hints.ai_next = NULL;
	error = getaddrinfo(NULL, servname, &hints, &res0);
	if (error) {
		syslog(LOG_ERR, "%s", gai_strerror(error));
		return 1;
	}
	for (res = res0, nsock = 0; res && nsock < nelems(s); res = res->ai_next) {
		s[nsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s[nsock] < 0) {
			cause = "socket";
			continue;
		}

		if (bind(s[nsock], res->ai_addr, res->ai_addrlen) < 0) {
			cause = "bind";
			close(s[nsock]);
			continue;
		}
		(void) listen(s[nsock++], BACKLOG);
	}
	if (nsock == 0) {
		syslog(LOG_ERR, "%s", cause);
		return 1;
	}
	freeaddrinfo(res0);

	to = 0;
	for (;;) {
		int i;
		int served;

		FD_ZERO(&readfds);

		nfds = 0;
		for (i = 0; i < nsock; i++) {
			FD_SET(s[i], &readfds);
			if (s[i] > nfds) {
				nfds = s[i];
			}
		}

		nfds++;

		if (to == 0) {
			tos = NULL;
		}
		else {
			timeout.tv_sec = to / 1000;
			timeout.tv_usec = (to % 1000) * 1000;
			tos = &timeout;
		}

		syslog(LOG_DEBUG, "| selecting\n");
/* XXX timeout ga 2 kai izyou tuduitara, motto nagaku site yoi */
		if (select(nfds, &readfds, NULL, NULL, tos) <= 0) {
			to = (*serv)(-1, data, NULL);
			continue;
		}

		served = 0;
		for (i = 0; i < nsock; i++) {
			struct sockaddr_storage addr;
			socklen_t len;
			u_int serv_to;
			syslog(LOG_DEBUG, "| accepting\n");
			if (!FD_ISSET(s[i], &readfds)) {
				continue;
			}
			len = sizeof addr;
			if ((ss = accept(s[i], (struct sockaddr *)&addr, &len)) < 0) {
				perror("accept");
				continue;
			}
			syslog(LOG_DEBUG, "| accepted\n");
			serv_to = (*serv)(ss, data, &addr);
			if (!served) {
				to = serv_to;
				served = 1;
			}
			else {
				to = MIN(to, serv_to);
			}
		}
	}
	return 0;
}

int
sockets(servname, s, size, cause)
char const *servname;
int s[];
size_t size;
const char **cause;
{
	int nsock;
	struct addrinfo hints, *res, *res0;
	int e;
	static char cs[1024];

	*cause = NULL;

	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_protocol = 0;
	hints.ai_addrlen = 0;
	hints.ai_canonname = NULL;
	hints.ai_addr = NULL;
	hints.ai_next = NULL;
	if (e = getaddrinfo(NULL, servname, &hints, &res0)) {
		syslog(LOG_ERR, "%s", gai_strerror(e));
		return 0;
	}

	for (res = res0, nsock = 0; res && nsock < size; res = res->ai_next) {
		s[nsock] = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if (s[nsock] < 0) {
			snprintf(cs, sizeof cs, "socket: %s", strerror(errno));
			*cause = cs;
			continue;
		}

		if (bind(s[nsock], res->ai_addr, res->ai_addrlen) < 0) {
			snprintf(cs, sizeof cs, "bind: %s", strerror(errno));
			*cause = cs;
			close(s[nsock]);
			continue;
		}
		(void) listen(s[nsock++], BACKLOG);
	}
	freeaddrinfo(res0);
	return nsock;
}
