/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: pkt.c,v 1.11 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>
#include <geta/timer.h>

#include "pkt.h"
#include "iov.h"
#include "dwcP.h"

#define	RECV	recv
#define	WRITEV	writev
/* #define	WRITEV	safe_writev */

static ssize_t recvl(int, void *, size_t, int);
#if defined XXDEBUG
static char const *cmdname(int);
static int seq = 0;
#endif

static struct {
	size_t send, recv;
	size_t scnt, rcnt, rcnt_sys;
	size_t smax, rmax;
	struct timeval stim, rtim;
} tc = {0, 0, 0, 0, 0, 0, 0, {0, 0}, {0, 0}};

ssize_t
send_pkt_c(s, sid, cmd, data, len)
void const *data;
size_t len;
{
	struct msg m;
	struct iovec iov[2];
	size_t iovcnt;
	ssize_t e;
	struct timeval start, end;

#if defined XXDEBUG
	m.h.seq = htonl(seq);
fprintf(stderr, "%s: send_pkt_c: socket=%d, seq=%d, sid=%d, cmd=%s, data=%p, len=%s\n", udate(), s, seq, sid, cmdname(cmd), data, sizetostr(len));
	seq++;
#endif

	if (len > MAXPKTLEN - sizeof m.h) {
		fprintf(stderr, "%s: send_pkt_c: too large data cmd=%d len=%s\n", udate(), cmd, sizetostr(len));
		return -1;
	}

	gettimeofday(&start, NULL);
	m.h.cmd = htonl(cmd);
	m.h.len = htonl(len);
	m.h.session_id = htonl(sid);
	iov[0].iov_base = (char *)&m.h;
	iov[0].iov_len = sizeof m.h;
	if (len > 0) {
		iov[1].iov_base = (void *)data;
		iov[1].iov_len = len;
		iovcnt = 2;
	}
	else {
		iovcnt = 1;
	}

	if ((e = WRITEV(s, iov, iovcnt))==-1) {
		dwerrno = errno;
fprintf(stderr, "%s: writev failed\n", udate());
		return -1;
	}
	tc.send+=e;
	tc.scnt++;
	if (tc.smax<e) {
		tc.smax = e;
	}
	gettimeofday(&end, NULL);
	timersub(&end, &start, &end);
	timeradd(&end, &tc.stim, &tc.stim);
#if defined XXDEBUG
fprintf(stderr, "%s: send_pkt_c: %s (%s+%s) done.\n", udate(), sizetostr(e), sizetostr(sizeof m.h), sizetostr(len));
#endif
	return e;
}

ssize_t
send_pkt_v(s, sid, cmd, iov, iovcnt)
struct iovec iov[];
size_t iovcnt;
{
	struct msg m;
	struct iovec iovP[32];
	size_t iovcntP;
	size_t len;
	size_t i;
	ssize_t e;
	struct timeval start, end;

#if defined XXDEBUG
	m.h.seq = htonl(seq);
fprintf(stderr, "%s: send_pkt_v: socket=%d, seq=%d, sid=%d, cmd=%s, iov=%p, iovcnt=%s\n", udate(), s, seq, sid, cmdname(cmd), iov, sizetostr(iovcnt));
	seq++;
#endif

	gettimeofday(&start, NULL);
	iovcntP = iovcnt+1;
	if (iovcntP > nelems(iovP)) {
		dwerrno = DWE_LIMIT;
fprintf(stderr, "%s: buffer overrun\n", udate());
		return -1;
	}
	len = 0;
	for (i=0; i<iovcnt; i++) {
		iovP[i+1].iov_base = iov[i].iov_base;
		iovP[i+1].iov_len = iov[i].iov_len;
		len += iov[i].iov_len;
	}

	if (len > MAXPKTLEN - sizeof m.h) {
		fprintf(stderr, "%s: send_pkt_c: too large data cmd=%d len=%s\n", udate(), cmd, sizetostr(len));
		return -1;
	}

	m.h.cmd = htonl(cmd);
	m.h.len = htonl(len);
	m.h.session_id = htonl(sid);
	iovP[0].iov_base = (char *)&m.h;
	iovP[0].iov_len = sizeof m.h;

#if defined XXDEBUG
fprintf(stderr, "%s: s:%d iovP:%p iovcntP:%s\n", udate(), s, iovP, sizetostr(iovcntP));
#endif
	if ((e = WRITEV(s, iovP, iovcntP))==-1) {
		dwerrno = errno;
fprintf(stderr, "%s: writev failed\n", udate());
		return -1;
	}
	tc.send+=e;
	tc.scnt++;
	if (tc.smax<e) {
		tc.smax = e;
	}
	gettimeofday(&end, NULL);
	timersub(&end, &start, &end);
	timeradd(&end, &tc.stim, &tc.stim);
#if defined XXDEBUG
fprintf(stderr, "%s: send_pkt_v: %s (%s+%s) done.\n", udate(), sizetostr(e), sizetostr(sizeof m.h), sizetostr(len));
#endif
	return e;
}

int
recv_pkt_c(s, cmd, m)
struct msg *m;
{
	if (recv_pkt(s, m)==-1) {
#if 1
		fprintf(stderr, "%s: recv_pkt_c failed: while waiting for %x !recv_pkt\n", udate(), cmd);
#endif
		return -1;
	}
	if (m->h.cmd != cmd) {
#if 1
		fprintf(stderr, "%s: recv_pkt_c failed: cmd mismatch: %d!=%d, len=%s\n", udate(), m->h.cmd, cmd, sizetostr(m->h.len));
#endif
		dwerrno = DWE_MISM;
		return -1;
	}
	return 0;
}

int
recv_pkt(s, m)
struct msg *m;
{
	ssize_t e;
	int flags = 0;
	struct timeval start, end;

errno = 0;
	gettimeofday(&start, NULL);
	if ((e = recvl(s, &m->h, sizeof m->h, flags)) != sizeof m->h) {
		struct sockaddr_in peeraddr;
#if ! HAVE_SOCKLEN_T
#if ! defined(socklen_t)
#define	socklen_t size_t
#endif
#endif
		socklen_t namelen = sizeof peeraddr;
		dwerrno = errno;
#if 1
fprintf(stderr, "%s: recv_pkt: recvl(1) failed: e=%s (%s)\n", udate(), ssizetostr(e), strerror(errno));
#endif
		getpeername(s, (struct sockaddr *)&peeraddr, &namelen);
#if 0
/* XXX inet_ntoa		fprintf(stderr, "%s: peer=%s\n", udate(), inet_ntoa(peeraddr.sin_addr)); */
#endif
		return -1;
	}
	m->h.cmd = ntohl(m->h.cmd);
	m->h.len = ntohl(m->h.len);
	m->h.session_id = ntohl(m->h.session_id);
#if defined XXDEBUG
	m->h.seq = ntohl(m->h.seq);
#endif
	tc.recv+=e;
	tc.rcnt++;
	if (tc.rmax<e) {
		tc.rmax = e;
	}
#if defined XXDEBUG
fprintf(stderr, "%s: recv_pkt: socket=%d, seq=%d, sid=?, m=%p, cmd=%s, e=%s, len=%s\n", udate(), s, m->h.seq, m, cmdname(m->h.cmd), sizetostr(e), sizetostr(m->h.len));
#endif

	if (m->h.len > 0) {
#if defined XXDEBUG
fprintf(stderr, "%s: recv_pkt: (data): data=%p %s\n", udate(), m->u.data, sizetostr(e));
#endif
		if (m->h.len > m->u.size) {
			fprintf(stderr, "%s: too large data: %s > %s, cmd=%s\n", udate(), sizetostr(m->h.len), sizetostr(m->u.size), sizetostr(m->h.cmd));
			dwerrno = DWE_LIMIT;
			return -1;
		}
		if ((e = recvl(s, m->u.data, m->h.len, flags)) != m->h.len) {
#if 1
			fprintf(stderr, "%s: recv_pkt: recvl(2) failed: e=%s (%s)\n", udate(), ssizetostr(e), strerror(errno));
#endif
			dwerrno = errno;
			return -1;
		}
		tc.recv+=e;
		tc.rcnt++;
		if (tc.rmax<e) {
			tc.rmax = e;
		}
	}
	gettimeofday(&end, NULL);
	timersub(&end, &start, &end);
	timeradd(&end, &tc.rtim, &tc.rtim);

#if defined XXDEBUG
fprintf(stderr, "%s: recv_pkt: done.\n", udate());
#endif

	return 0;
}

static ssize_t
recvl(s, buf, len, flags)
void *buf;
size_t len;
{
	ssize_t e, c;
	int ecnt;
	int tout_sec = 0, tout_msec = 1;
#define	ELIM	20

	for (c = 0, ecnt = -3; c < len && ecnt < ELIM; c += e) {
		if (0 < ecnt) {
			fd_set readfds, *rfds = NULL;
			struct timeval timeout;
			timeout.tv_sec = tout_sec;
			timeout.tv_usec = tout_msec * 1000;
			if (ecnt < 5) {
				FD_ZERO(&readfds);
				FD_SET(s, &readfds);
				rfds = &readfds;
			}
/*#if defined XXDEBUG*/
#if 1
if (!rfds) {
	fprintf(stderr, "%s: recv_pkt: sleep for %d.%03dsec...\n", udate(), tout_sec, tout_msec);
}
#endif
			if (select(s + 1, rfds, NULL, NULL, &timeout) == -1) {
				perror("select");
				return -1;
			}
			if (ecnt < 9) {
				tout_msec *= 2;
			}
			else {
				tout_sec = 1;
				tout_msec = 0;
			}
		}
		assert(len - c != 0);
		if ((e = RECV(s, (char *)buf + c, len - c, flags)) == -1) {
			dwerrno = errno;
			fprintf(stderr, "%s: recvl: RECV failed (%s)\n", udate(), strerror(errno));
			return -1;
		}
		tc.rcnt_sys++;
		if (e == 0) {
			ecnt++;
		}
	}

	if (c < len) {
		fprintf(stderr, "%s: recvl: failed: short packet: s=%d, c=%s\n", udate(), s, sizetostr(c));
		return -1;
	}

	if (c > len) {
		fprintf(stderr, "%s: recvl: failed: overrun!\n", udate());
		dwerrno = DWE_LIMIT;
		return -1;
	}

	return c;
}

void
pkt_d(s)
FILE *s;
{
	fprintf(s, "(send %s) (scnt %s) (smax %s)", 
		sizetostr(tc.send), sizetostr(tc.scnt), sizetostr(tc.smax));
	fprintf(s, " ");
	fprintf(s, "(recv %s) (rcnt %s) (rmax %s) (rcnt_sys %s)", 
		sizetostr(tc.recv), sizetostr(tc.rcnt), sizetostr(tc.rmax), sizetostr(tc.rcnt_sys));
#define	PPP(stream, name, var) \
	(fprintf((stream), " (" #name " "), \
	 timerprint((stream), (var)), \
	 fprintf((stream), ")"))

	PPP(s, stim, &tc.stim);
	PPP(s, rtim, &tc.rtim);
#undef PPP
}

#if defined XXDEBUG
static char const *
cmdname(cmd)
{
	static char s[32];
#define	foo(n)	{ #n, DWCC_##n }
static struct { char const *name; int cmd; } cmds[] = {
	foo(HND), foo(BFS), foo(OPN), foo(CRE),
	foo(CLO), foo(VEC), foo(FLU), foo(WVE),
	foo(SVI), foo(SVR), foo(ATV), foo(BLK),
	foo(SBI), foo(SBR), foo(WSH), foo(WSHS),
	foo(RQN), foo(NDL), foo(SYN), foo(CM),
	foo(MSK), foo(DIS),
};
#undef foo
	static char const *rcmds[2 * nelems(cmds) + 2];
	static int initialized = 0;
	if (!initialized) {
		size_t i;
		for (i = 0; i < nelems(rcmds); i++) {
			rcmds[i] = NULL;
		}
		for (i = 0; i < nelems(cmds); i++) {
			int c = cmds[i].cmd;
			if (c < nelems(rcmds)) {
				rcmds[c] = cmds[i].name;
			}
			c |= DWCC_ACK;
			if (c < nelems(rcmds)) {
				snprintf(s, sizeof s, "%s+", cmds[i].name);
				rcmds[c] = strdup(s);
			}
		}
		initialized = 1;
	}
	if (0 <= cmd && cmd < nelems(rcmds) && rcmds[cmd]) {
		return rcmds[cmd];
	}
	snprintf(s, sizeof s, "%d%s", cmd>>1, cmd&1?"+":"");
	return s;
}
#endif
