/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: sched.c,v 1.6 2007/03/29 02:10:01 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <assert.h>
#if defined HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#if defined HAVE_MACH_O_DYLD_H
#include <mach-o/dyld.h>
#endif
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#if defined HAVE_LINK_H
#include <link.h>
#endif
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>

#include "dwsys.h"
#include "dxrP.h"
#include "iosched.h"
#include "csche.h"
#include "sdup.h"
#include "nht.h"

static int load_config(struct iosched *, char const *, char const *);
static int lddll(struct iosched *, char *);
static int dcompar(void const *, void const *);
static int tcompar(void const *, void const *);
static int sc_init_vec(struct iosched *, struct sched *, size_t, int);
static void count_elem_num(struct sched *, struct sched *);
static int mkpath(struct iosched *, struct sched_iocb *, int, ssize_t);
static ssize_t selp_ileave(struct sched *, idx_t);
static mkrhost_t mkrhost_Localhost;

static void print_local(FILE *, struct sched_iocb *);
static void print_remote(FILE *, struct sched_iocb *);

static int resolv_dll(void *, const char *, const char *, void (**)(void));
static int resolv_hw(struct hwfl *, const char *, const char *, void (**)(void));

/*
 * schedp の文法:
 * #row-partitions . #col-partitions . type [. arg-to-custom]
 * ここに, 
 * #x-partitions: x を分割する数
 * type: スケジュールタイプ. sched.conf の対応する定義がロードされる.
 * arg-to-custom はそのまま custom に渡される. 省略した場合, 
 * 空文字列 (NULL ではない)が渡される.
 *
 * sched.conf の文法
 * type : path : mkrhost : custom : pretty-name
 * ここに, 
 * path: mkrhost, custom を含むダイナミックライブラリのパス名
 *   * と書くと, 何もしない (mkrhost, custom は 無視される.)
 *   - と書くと, ハードワイアード
 * mkrhost: 行, 列の別と, パーティション番号を渡されて, そのパーティション
 *  を担当するホストの名前を返す関数.
 * custom: 実際にスケジューラをカスタマイズする関数.
 *  主として, selpp などをセットする.
 *
 * mkrhost, custom に default と記せば, 既定値が用いられる.
 *
 * selpp:  行, 列の別と, 記事や単語のIDを渡されて, 
 *  その記事や単語を入れるパーティションの番号を返す関数.
 *
 * libcsche.so の指定方法
 * LD_LIBRARY_PATH=$HOME/exp/geta/lib:$LD_LIBRARY_PATH 
 * export LD_LIBRARY_PATH
 * 
 */

#define	DEFAULT	"default"
#define NODLL	"*"
#define HWIRED	"-"

int
schedule_set_policy(s, schedp, sched_conf)
struct iosched *s;
char *schedp;
char const *sched_conf;
{
	char *p, *end;

	p = schedp;
	s->scc[0].np = s->scc[1].onp = strtosize(p, &end, 10);
	if (p==end || *end != '.') {
		fprintf(stderr, "set_policy: parse error: 1\n");
		return -1;
	}
	p = end+1;
	s->scc[1].np = s->scc[0].onp = strtosize(p, &end, 10);
	if (p==end || *end != '.') {
		fprintf(stderr, "set_policy: parse error: 2\n");
		return -1;
	}
	p = end+1;

	s->type = p;

	for (; *p && *p!='.'; p++) ;
	if (*p=='.') {
		*p++ = '\0';
	}

	s->param = p;
	s->custom = NULL;
	s->mkrhost = mkrhost_Localhost;

	if (load_config(s, s->type, sched_conf)==-1) {
		fprintf(stderr, "set_policy failed: 3\n");
		return -1;
	}
	return 0;
}

static int
load_config(s, type, sched_conf)
struct iosched *s;
char const *type, *sched_conf;
{
	FILE *f;
	static char buf[1024];
	size_t l;

	printf("load_config: loading %s\n", type);

	if (!(f = fopen(sched_conf, "r"))) {
		nperror(sched_conf);
		return -1;
	}

	l = strlen(type);
	while (fgets(buf, sizeof buf, f)) {
		char *p;
		if (p = rindex(buf, '\n')) {
			*p = '\0';
		}
		if (!strncmp(buf, type, l) && buf[l] == ':') {
			fclose(f);
			return lddll(s, buf+l+1);
		}
	}
	fclose(f);

	fprintf(stderr, "load_sched: error: [%s] not found in [%s]\n", type, sched_conf);

	return -1;
}

static int
resolv_dll(dl, name, pname, p)
void *dl;
const char *name, *pname;
void (**p)(void);
{
	if (strcmp(name, DEFAULT)) {	/* unless DEFAULT */
		void (*v)(void);
#if defined HAVE_MACH_O_DYLD_H
		_dyld_lookup_and_bind(name, (unsigned long *)&v, NULL);

#else
#if HAVE_DLFCN_H
		if (!(v = dlsym(dl, name))) {
			fprintf(stderr, "\n%s: %s\n", name, dlerror());
			return -1;
		}
#endif
#endif
		*p = v;
		printf(", %s = %p", pname, *p);
	}
	else {
		printf("%s", ", " DEFAULT);
	}
	return 0;
}

static int
resolv_hw(fl, name, pname, p)
struct hwfl *fl;
const char *name, *pname;
void (**p)(void);
{
	if (strcmp(name, DEFAULT)) {	/* unless DEFAULT */
		void (*v)(void);
		size_t i;
		v = NULL;
		for (i=0; fl[i].name; i++) {
			if (!strcmp(name, fl[i].name)) {
				v = fl[i].func;
				break;
			}
		}
		if (v == NULL) {
			fprintf(stderr, "\n%s: not configured.\n", name);
			return -1;
		}
		*p = v;
		printf(", %s = %p", pname, *p);
	}
	else {
		printf("%s", ", " DEFAULT);
	}
	return 0;
}

static struct hwfl hwfl_mkrhosts[] = {
	HWFL_MKRHOSTS
	{NULL, NULL}};

static struct hwfl hwfl_customs[] = {
	HWFL_CUSTOMS
	{NULL, NULL}};

static int
lddll(s, buf)
struct iosched *s;
char *buf;
{
	char const *path, *mkrhost, *custom, *pretty_name;
	char *p;

	p = buf;

	path = p;
	if (!(p = index(p, ':'))) {
		fprintf(stderr, "lddll: parse error: [%s]\n", buf);
		return -1;
	}
	*p++ = '\0';

	mkrhost = p;
	if (!(p = index(p, ':'))) {
		fprintf(stderr, "lddll: parse error: [%s]\n", buf);
		return -1;
	}
	*p++ = '\0';

	custom = p;
	if (!(p = index(p, ':'))) {
		fprintf(stderr, "lddll: parse error: [%s]\n", buf);
		return -1;
	}
	*p++ = '\0';

	pretty_name = p;

	printf("%s, %s, %s\n", sizetostr(s->scc[0].np), sizetostr(s->scc[1].np), pretty_name);

	printf("%s", path);

	if (!strcmp(path, HWIRED)) {
#if defined HW_CSCHE
		if (resolv_hw(hwfl_mkrhosts, mkrhost, "mkrhost", (void(**)(void))&s->mkrhost)==-1) {
			return -1;
		}
		if (resolv_hw(hwfl_customs, custom, "custom", (void(**)(void))&s->custom)==-1) {
			return -1;
		}
#else
		fprintf(stderr, "lddll: csche is not hardwired.\n");
		return -1;
#endif
	}
	else if (!strcmp(path, NODLL)) {
		/* do nothing */
	}
	else {
		void *dl;
#if defined HAVE_MACH_O_DYLD_H
		if (_dyld_present()==0) {
			perror("_dyld_present");
			return -1;
		}
#else
#if HAVE_DLFCN_H
		if (!(dl = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
			fprintf(stderr, "\n%s: %s\n", path, dlerror());
			return -1;
		}

/*
 *** process syuuryo made dlclose siteha ikenai.
		dlclose(dl);
*/
#else
		fprintf(stderr, "lddll: dynamic linker is not configured.\n");
		return -1;
#endif
#endif

		if (resolv_dll(dl, mkrhost, "mkrhost", (void(**)(void))&s->mkrhost)==-1) {
			return -1;
		}
		if (resolv_dll(dl, custom, "custom", (void(**)(void))&s->custom)==-1) {
			return -1;
		}
	}

	printf("\n");

	return 0;
}

static void
print_local(stream, iocb)
FILE *stream;
struct sched_iocb *iocb;
{
	fprintf(stream, "%s -", basename(iocb->L.path));
	fprintf(stream, " %s %s", freqtostr(iocb->total_elem_num), ssizetostr(iocb->elem_num));
	fprintf(stream, "\n");
}

static void
print_remote(stream, iocb)
FILE *stream;
struct sched_iocb *iocb;
{
	fprintf(stream, "%s", iocb->R.host);
	fprintf(stream, " %s", iocb->R.serv);
	fprintf(stream, " %s %s", freqtostr(iocb->total_elem_num), ssizetostr(iocb->elem_num));
	fprintf(stream, "\n");
}

int
print_sched(s, stream)
struct iosched *s;
FILE *stream;
{
	size_t p, d;

	fprintf(stream, "%c\n", s->local ? 'L' : 'N');
	fprintf(stream, "%s\n", basename(s->dcw_path[0]));
	fprintf(stream, "%s\n", basename(s->dcw_path[1]));
	for (d=0; d<=1; d++) {
		struct sched_iocb *iocb;
		if (s->local) {
			fprintf(stream, "%s\n", sizetostr(s->scc[d].np));
			iocb = &s->scc[d].iocb0;
			print_local(stream, iocb);
			for (p=0; p<s->scc[d].np; p++) {
				iocb = &s->scc[d].iocb[p];
				print_local(stream, iocb);
			}
		}
		else {
			fprintf(stream, "%s\n", sizetostr(s->scc[d].np));
			iocb = &s->scc[d].iocb0;
			print_remote(stream, iocb);
			for (p=0; p<s->scc[d].np; p++) {
				iocb = &s->scc[d].iocb[p];
				print_remote(stream, iocb);
			}
		}
	}

	return 0;
}

int
schedule(s, r, nr, c, nc, raw_max_elem_num)
struct iosched *s;
struct ht *r, *c;
size_t nr, nc;
ssize_t raw_max_elem_num;
{
	size_t i;
	ssize_t max_elem_num;

	s->local = 0;

	s->tg[0].h = r;
	s->tg[0].n = nr;
	s->tg[1].h = c;
	s->tg[1].n = nc;

#if 0
#if 0
	/* mukasi ha, hte_pcompar (name de hikaku) de sort siteita. 
	   aqs8k kouritu no men kara, DF de sort suru youni henkou. 
	   07/23/01 */
	qsort(r, nr, sizeof *r, hte_pcompar);
#else
#if HAVE_HEAPSORT
	if (heapsort(r, nr, sizeof *r, dcompar)==-1) {
		return -1;
	}
#else
	qsort(r, nr, sizeof *r, dcompar);
#endif
#endif
#else
	if (nht_sort(s->tg[0].h, s->tg[0].n, dcompar)==-1) {
		return -1;
	}
#endif
	max_elem_num = 0;
	for (i=0; i<s->tg[0].n; i++) {
		struct hte *e = hta(s->tg[0].h, i);
		e->i.id = i+1; /* XXX */
		if (max_elem_num < e->i.df) {
			max_elem_num = e->i.df;
		}
	}
	s->scc[0].max_elem_num = max_elem_num;
	s->scc[0].raw_max_elem_num = raw_max_elem_num;

#if 0
#if 0
	/* mukasi ha, tcompar (TF de hikaku) de sort siteita. 
	   aqs8k kouritu no men kara, DF de sort suru youni henkou. 
	   07/23/01 */
	qsort(c, nc, sizeof *c, tcompar);
#else
	qsort(c, nc, sizeof *c, dcompar);
#endif
#else
	if (nht_sort(s->tg[1].h, s->tg[1].n, dcompar)==-1) {
		return -1;
	}
#endif
	max_elem_num = 0;
	for (i=0; i<s->tg[1].n; i++) {
		struct hte *e = hta(s->tg[1].h, i);
		e->i.id = i+1; /* XXX */
		if (max_elem_num < e->i.df) {
			max_elem_num = e->i.df;
		}
	}
	s->scc[1].max_elem_num = max_elem_num;
	s->scc[1].raw_max_elem_num = 0;

	s->scc[0].selpp = s->scc[1].selpp = selp_ileave;

	s->scc[0].pp = NULL;
	s->scc[1].pp = NULL;

	if (s->custom && (*s->custom)(s, s->param)==-1) {
		return -1;
	}

	if (sc_init_vec(s, &s->scc[0], s->tg[0].n, DXR_ROW)==-1) {
		return -1;
	}
	if (sc_init_vec(s, &s->scc[1], s->tg[1].n, DXR_COL)==-1) {
		return -1;
	}

	/* count_elem_num ha, sc_init_vec ga syuuryou sita ato de nakereba yobidase nai. */
	count_elem_num(&s->scc[0], &s->scc[1]);
	count_elem_num(&s->scc[1], &s->scc[0]);

	return 0;
}

static int
dcompar(va, vb)
void const *va, *vb;
{
	struct hte const *a, *b;
	a = va;
	b = vb;
	/* descending order */
	if (a->i.df < b->i.df) {
		return 1;
	}
	else if (a->i.df > b->i.df) {
		return -1;
	}
	return strcmp(a->name.p, b->name.p);
}

static int
tcompar(va, vb)
void const *va, *vb;
{
	struct hte const *a, *b;
	a = va;
	b = vb;
	/* descending order */
	if (a->i.tf < b->i.tf) {
		return 1;
	}
	else if (a->i.tf > b->i.tf) {
		return -1;
	}
	return strcmp(a->name.p, b->name.p);
}

static int
sc_init_vec(s, sc, n, type)
struct iosched *s;
struct sched *sc;
size_t n;
{
	ssize_t p;
	sc->n = n;

	sc->iocb0.svec = NULL;
	sc->iocb0.total_elem_num = 0;
	sc->iocb0.elem_num = 0;
	if (mkpath(s, &sc->iocb0, type, -1)==-1) {
		return -1;
	}

	if (sc->np <= 0) {
		fprintf(stderr, "invalid np(%s)\n", sizetostr(sc->np));
		return -1;
	}
	if (!(sc->iocb = calloc(sc->np, sizeof *sc->iocb))) {
		nperror("calloc B");
		return -1;
	}
	sc->svec_size = sizeof_dxr_vec(sc->max_elem_num);
	for (p=0; p<sc->np; p++) {
		struct sched_iocb *iocb = &sc->iocb[p];
		iocb->svec = NULL;
		iocb->total_elem_num = 0;
		iocb->elem_num = 0;
		if (mkpath(s, iocb, type, p)==-1) {
			return -1;
		}
	}
	return 0;
}

static void
count_elem_num(sc0, sc1)
struct sched *sc0, *sc1;
{
	idx_t i;

	/* sc0->iocb[p].elem_num ha, sc_init_vec de zero ni syokika sareru */
	for (i = 1; i<=sc1->n; i++) {
		ssize_t p;
		p = (*sc1->selpp)(sc1, i);
		assert(0<=p);
		assert(p<sc0->np);
		sc0->iocb[p].elem_num++;
	}
	sc0->iocb0.elem_num = sc1->n;
}

static int
mkpath(s, iocb, type, fragment)
struct iosched *s;
struct sched_iocb *iocb;
ssize_t fragment;
{
	if (s->local) {
		return dw_mkpathname(iocb->L.path, sizeof iocb->L.path, s->dirname, s->handle, type, fragment, 0);
	}
	else {
		return (*s->mkrhost)(s, iocb, type, fragment);
	}
}

static ssize_t
selp_ileave(sc, id)
struct sched *sc;
idx_t id;
{
	return id % sc->onp;
}

/* mkrhost_Localhost ha, csche.c nimo sokkuri na monoga aru. 
 * (sotiraha, mkrhost_localhost toiu namae de aru. (l ga komozi))
 * default ga kotirano mkrhost_Localhost nanode,
 * custom de sotirano mkrhost_localhost wo sitei suru noto, 
 * mkrhost ni nanimo sitei sinai notoha onazi kekka ninaru.
 * sched.conf wo kaku toki wakari yasuuinode konomama ni sitearu 
 */ 
static int
mkrhost_Localhost(s, iocb, type, fragment)
struct iosched *s;
struct sched_iocb *iocb;
ssize_t fragment;
{
	iocb->R.host = "localhost";
#if defined DWD_PORT
	iocb->R.serv = DWD_PORT;
#else
	iocb->R.serv = DWD_SERV;
#endif
	return 0;
}
