/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: sexp.c,v 1.2 2007/03/21 22:34:27 nis Exp $"; */

#include <config.h>

#include <sys/types.h>
#include <sys/uio.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dwut.h>

#include "sexp.h"
#include "swam.h"

struct SEXP {
	int type;
	union {
		struct {
			struct SEXP *car, *cdr;
		} pair;
		char const *pname;
		int i;
		struct SEXP *next;	/* free cells chain */
	} value;
};

struct SEXP NIL = {'n'}, *nil = &NIL;
struct SEXP CLOSE_PAREN = {')'}, *close_paren = &CLOSE_PAREN;

static void release_cons(sexp);
static sexp get_cons(void);
static char *fullwd_alloc(size_t);

static char token[4096];

sexp
sexp_read()
{
	sexp r;
	int type = 0;

	do {
		type = gettoken(token, sizeof token);
	} while (type==' ' || type=='\n' || type==';');

	if (type==EOF) {
		return NULL;
	}
	else if (type=='(' || type=='v') {
		struct SEXP sentinel;
		sexp e, n, c;
		e = &sentinel;
		e->value.pair.cdr = nil;
		for (; (c = sexp_read())!=close_paren; ) {
			if (!c) {
				return NULL;
			}
			if (!(n = get_cons())) {
				return NULL;
			}
			n->type = 'p';
			n->value.pair.car = c;
			n->value.pair.cdr = nil;
			e->value.pair.cdr = n;
			e = n;
		}
		r = sentinel.value.pair.cdr;
	}
	else if (type=='"' || type=='t') {
		sexp n;
		char *p;
		size_t l;
		if (!(n = get_cons())) {
			return NULL;
		}
		n->type = type;
		l = strlen(token)+1;
		if (!(p = fullwd_alloc(l))) {
			return NULL;
		}
		bcopy(token, p, l);
		n->value.pname = p;
		r = n;
	}
	else if (type==')') {	/* ')' of parent */
		return close_paren;
	}
	else {	/* error! */
		return NULL;
	}

	return r;
}

void
sexp_print(s, stream)
sexp s;
FILE *stream;
{
	if (s==nil) {
		fprintf(stream, "()");
	}
	else if (s->type=='p') {	/* pair */
		fprintf(stream, "(");
		for (; s!=nil; s = s->value.pair.cdr) {
			if (s->type=='p') {
				sexp_print(s->value.pair.car, stream);
			}
			else {
				fprintf(stream, " . ");
				sexp_print(s, stream);
				break;
			}
			if (s->value.pair.cdr != nil) {
				fprintf(stream, " ");
			}
		}
		fprintf(stream, ")");
	}
	else if (s->type=='t') {
		fprintf(stream, "%s", s->value.pname);
	}
	else if (s->type=='"') {
		sexp_put_string(s->value.pname, stream);
	}
	else {
		fprintf(stderr, "error!\n");
	}
}

void
sexp_put_string(p, stream)
char const *p;
FILE *stream;
{
	char const *e, *q;
	size_t l;
	l = strlen(p);
	putc('"', stream);
	e = p+l;
	for (; p<e; p=q) {
		for (q=p; q<e && !(*q=='"' || *q=='\\'
#if defined JIS_WAM
				   || !strncmp(q, "\033$B", 3)
#endif
				   ); q++) ;
		fwrite(p, 1, q-p, stream);

		if (q<e) {
			if ((*q=='"' || *q=='\\')) {
				putc('\\', stream);
				putc(*q, stream);
				q++;
			}
#if defined JIS_WAM
			else if (!strncmp(q, "\033$B", 3)) {
				char const *r, *s;
				for (r=q; r<e && strncmp(r, "\033(B", 3); r++) ;
				if (r+3<=e) {
					r += 3;
				}
				else {
					fprintf(stderr, "?");
				}
				for (s=q+3; s<r-3; s++) {
					putc(*s|0x80, stream);
				}
				q = r;
			}
#endif
			else {
				fprintf(stderr, "never happen\n");
			}
		}
	}
	putc('"', stream);
}

char const *
symbol_name(s)
sexp s;
{
	if (s->type=='t' || s->type=='"') {
		return s->value.pname;
	}
	return NULL;
}

size_t
length(s)
sexp s;
{
	size_t n;
	for (n=0; s->type=='p'; n++) {
		s = s->value.pair.cdr;
	}
	if (s==nil) {
		return n;
	}
	return -1;
}

int
pairp(s)
sexp s;
{
	return s->type=='p';
}

sexp
car(s)
sexp s;
{
	if (s->type=='p') {
		return s->value.pair.car;
	}
	return nil;
}

sexp
cdr(s)
sexp s;
{
	if (s->type=='p') {
		return s->value.pair.cdr;
	}
	return nil;
}

sexp
nth(s, n)
sexp s;
{
	for (; s->type=='p' && 0<n; n--) {
		s = s->value.pair.cdr;
	}
	if (s->type=='p') {
		return s->value.pair.car;
	}
	return nil;
}

static struct SEXP *free_cells = NULL;

static struct SEXP *cell_pool = NULL;
static size_t nfree = 0;

#define	NPOOLS	8192
static struct SEXP *pools[NPOOLS];
static size_t npools = 0, lastnpools = 0;

static char *fullwd_frags = NULL;
static size_t fullp = 0;

#define	NFFRGS	8192
static char *ffrgs[NFFRGS];
static size_t nffrgs = 0, lastnffrgs = 0;

static void
release_cons(c)
sexp c;
{
	c->type = 'f';
	c->value.next = free_cells;
	free_cells = c;
}

static sexp
get_cons()
{
#define	NBULK	8192
	if (free_cells) {
		sexp r;
		assert(free_cells->type == 'f');
		r = free_cells;
		free_cells = free_cells->value.next;
		return r;
	}
	if (nfree==0) {
		if (nelems(pools) <= npools) {
			return NULL;
		}
		if (lastnpools <= npools) {
			if (!(pools[npools] = calloc(sizeof *pools[npools], NBULK))) {
				return NULL;
			}
			lastnpools = npools + 1;
		}
		cell_pool = pools[npools++];
		nfree = NBULK;
	}
	nfree--;
	return &cell_pool[nfree];
}

static char *
fullwd_alloc(n)
size_t n;
{
#define	FFRGSZ	16384
	if (FFRGSZ < n) {
		/* bug: FFRGSZ yori ookii moziretu wo okuru kotoha dekinai.
		   koreha, wstem ni totte, kitui seiyaku de aru. */
fprintf(stderr, "<");
		return NULL;
	}
	if (fullp < n) {
		if (nelems(ffrgs) <= nffrgs) {
			return NULL;
		}
		if (lastnffrgs <= nffrgs) {
			if (!(ffrgs[nffrgs] = malloc(FFRGSZ))) {
				return NULL;
			}
			lastnffrgs = nffrgs + 1;
		}
		fullwd_frags = ffrgs[nffrgs++];
		fullp = FFRGSZ;
	}
	fullp -= n;
	return &fullwd_frags[fullp];
}

/* relase all cons-cell and fullword */
void
relall()
{
	nfree = npools = 0;
	fullp = nffrgs = 0;
}
